"""
This package contains functions for modifying bounding boxes
"""

# make functions available at the package level using shadow imports
from .transforms import clip_boxes
from .transforms import tf_box_transform
from .transforms import parameterize
from .transforms import unparameterize
from .transforms import filter_edge_boxes
from .neighborhood import assemble_single_neighborhood, all_neighborhoods_additional_info
from .match import cluster_assignment

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'clip_boxes',
    'tf_box_transform',
    'parameterize',
    'unparameterize',
    'filter_edge_boxes',
    'assemble_single_neighborhood',
    'all_neighborhoods_additional_info',
    'cluster_assignment'
)



"""
The networks package contains functions for creating and manipulating networks
"""

# make functions available at the package level using shadow imports
from .backbones import residual, pretrained
from .fast_rcnn import fast_rcnn
from .field_size import field_size
from .rpns import rpn
from .transfer_layers import transfer_layers

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'residual',
    'pretrained',
    'fast_rcnn',
    'field_size',
    'rpn',
    'transfer_layers'
)



"""
This package contains functions for loading and saving data and results
"""

# make functions available at the package level using shadow imports
from .input import dataset
from .input import read_csv
from .input import read_png
from .input import resize


# list out things that are available for public use
__all__ = (
    'dataset',
    'read_csv',
    'read_png',
    'resize'
)



import tensorflow as tf


def fast_rcnn(features, units=[4096, 4096], activations=['relu', 'relu'], pool=2, tiles=3):
    """
    Produces the terminal network used to refine regressions from the region-proposal network
    and to optionally classify the objects.
    
    Parameters
    ----------
    features : int
        The number of features in the 3D feature map generated by the backbone. Required to
        set input shape properties of network.
    units : array_like (int)
        The number of units in each dense layer. Default value [4096, 4096].
    activations : array_like (string)
        A list of strings defining the activations to use in the dense layers.
        Must be same length as dense. Default value ['relu', 'relu'].    
    pool : int
        pool^2 is the number of locations to interpolate features at within 
        each tile. Default value 2.
    tiles : int
        tile^2 is the number of tiles that each regressed bounding box is divided
        into. Default value 3.

    Returns
    -------
    fastrcnn : tf.keras.Model
        .
    """
    
    #generates fast RCNN model used in second regression that follows region proposal network.
    
    #create input layer
    fastrcnn_input = tf.keras.Input(shape=(pool*tiles, pool*tiles, features))

    #pooling over each roialign tile
    pooled = tf.keras.layers.MaxPool2D(pool_size=(pool, pool), padding='valid')(fastrcnn_input)

    #stack the pooled feature vectors from each tile into a single feature vector
    pooled = tf.reshape(pooled, (tf.shape(pooled)[0], tiles*tiles*features))

    #fully connected layers
    for i, (unit, activation) in enumerate(zip(units, activations)):
        if i == 0:
            layer_input = pooled
        else:
            layer_input = x
        x = tf.keras.layers.Dense(unit, activation=activation)(layer_input)

    #final layer to generate parameterized regression prediction
    regression_align = tf.keras.layers.Dense(4, activation='linear')(x)
    
    #create model
    fastrcnn = tf.keras.Model(inputs=fastrcnn_input,
                              outputs=[regression_align])
    
    return fastrcnn



from typing import List, Union

import tensorflow as tf
import numpy as np
import json

from histomics_detect.models.lnms_model import LearningNMS
from histomics_detect.anchors.create import create_anchors
from histomics_detect.metrics.iou import iou, greedy_iou_mapping
from histomics_detect.boxes.transforms import filter_edge_boxes
from histomics_detect.models.compression_network import CompressionNetwork


def save_history(history_callback, model: tf.keras.Model, experiment_id: int,
                 history_path: str = "JsonLogs/logs_{}.json", weight_path: str = "lnms_weights/cpk_{}") -> None:
    with open(history_path.format(experiment_id), 'w') as file:
        json.dump(history_callback.history, file)

    model.save_weights(weight_path.format(experiment_id))


def validate_model(ds_validation_roi: tf.data.Dataset, model: LearningNMS, faster_model, experiment_id: int = 0,
                   path: str = "JsonLogs/logs_{}_auc.json"):
    tp_counter, fp_counter, fn_counter = np.float32(0), np.float32(0), np.float32(0)
    aucs = dict()
    average_auc = np.float32(0)
    counter = np.float32(0)

    for data in ds_validation_roi:
        (rgb, boxes, name) = data
        boxes = boxes.to_tensor()
        width, height, _ = tf.shape(rgb)

        model.anchors = create_anchors(model.anchor_px, model.field, height, width)

        features, rpn_boxes, scores, nms_output = model(data)

        filtered_boxes = tf.squeeze(
            tf.gather(rpn_boxes, tf.where(tf.greater(tf.reshape(nms_output, -1), 0.5))))
        if tf.size(filtered_boxes) > 0:
            try:
                rpn_boxes_final = faster_model.align(filtered_boxes, features, faster_model.field, faster_model.pool,
                                                     faster_model.tiles)

                boxes = filter_edge_boxes(boxes, tf.shape(rgb)[1], tf.shape(rgb)[0], 32,
                                          tf.constant(False, tf.bool))

                rpn_boxes_final, condition = filter_edge_boxes(rpn_boxes_final, tf.shape(rgb)[1], tf.shape(rgb)[0], 32,
                                                               tf.constant(False, tf.bool))
                scores = tf.reshape(tf.gather(scores, tf.where(condition)), (-1, 1))

                if tf.size(rpn_boxes_final) > 0:
                    ious = iou(rpn_boxes_final, boxes)

                    tp, fp, fn, tp_list, fp_list, fn_list = greedy_iou_mapping(ious, 0.18)
                    tp, fp, fn = tp.numpy(), fp.numpy(), fn.numpy()
                else:
                    tp, fp, fn = 0, 0, tf.shape(boxes)[0].numpy()
            except:
                tp, fp, fn = 0, 0, tf.shape(boxes)[0].numpy()

            try:
                scores = tf.reshape(scores, -1)
                obj = tf.stack([1 - scores, scores], axis=1)
                # auc = greedy_pr_auc(obj, rpn_boxes_final, boxes, delta=0.1, min_iou=0.18)
                # aucs[str(name.numpy())] = auc.numpy().item()
                # average_auc += auc.numpy()
            except:
                aucs[str(name.numpy())] = 0
        else:
            tp, fp, fn = 0, 0, tf.shape(boxes)[0].numpy()
            aucs[str(name.numpy())] = 0

        tp_counter += tp
        fp_counter += fp
        fn_counter += fn
        counter += 1

    log = {
        'counter': counter,
        'fp': fp_counter.item(),
        'tp': tp_counter.item(),
        "fn": fn_counter.item(),
        # 'aucs': aucs,
        # 'auc': average_auc.item(),
        # 'avg_auc': average_auc.item() / counter
    }

    for key, value in log.items():
        print(key, value, type(value))

    for key, value in log['aucs'].items():
        print(str(key), value, type(value))

    print(log)

    with open(path.format(experiment_id), 'w') as file:
        json.dump(log, file)


def run_experiments(ds_train_roi, ds_validation_roi, callbacks: List[tf.keras.callbacks.Callback], configs: dict,
                    faster_model: tf.keras.Model, optimizer: tf.keras.optimizers.Optimizer,
                    changing_variable: str, values: List[Union[str, tf.keras.losses.Loss, int, float, tf.Tensor]],
                    experiment_start_id: int = 0, epochs: int = 100, steps_per_epoch: int = 92) -> None:
    experiment_id = experiment_start_id
    for value in values:
        temp_configs = configs.copy()
        temp_configs[changing_variable] = value

        compression_net = CompressionNetwork(temp_configs['feature_size'], temp_configs['anchor_size'],
                                             faster_model.backbone)

        model = LearningNMS(configs, faster_model.rpnetwork, faster_model.backbone, compression_net.compression_layers,
                            [temp_configs['width'], temp_configs['height']], )
        model.compile(optimizer=optimizer)

        try:
            history_callback = model.fit(x=ds_train_roi, batch_size=1, epochs=epochs, verbose=1,
                                         callbacks=callbacks, steps_per_epoch=steps_per_epoch)  # 92
            try:
                save_history(history_callback, model, experiment_id)
                validate_model(ds_validation_roi, model, faster_model, experiment_id)
            except:
                print("validation or save failed")
        except:
            print("training failed")

        experiment_id += 1



import tensorflow as tf

from histomics_detect.models.faster_rcnn import map_outputs
from histomics_detect.boxes.transforms import unparameterize


def extract_data(data):
    """
    extracts image and boxes from the data

    Parameters
    ----------
    data: Tuple
        the image, the ground truth boxes, and the name of the image

    Returns
    -------
    norm: tensor (float32)
        normalized image
    boxes: tensor (float32)
        shape: D x 4
        ground truth boxes for the given image
    image_name: string
        name of the image
    """
    if len(data) == 3:
        rgb, boxes, image_name = data
    else:
        rgb, boxes = data
        image_name = None

    rgb = tf.squeeze(rgb)

    # convert boxes from RaggedTensor
    expand_fn = lambda x: tf.expand_dims(x, axis=0)
    boxes = boxes.to_tensor()
    boxes = tf.squeeze(boxes)
    boxes = tf.cond(tf.size(tf.shape(boxes)) == 1, lambda: expand_fn(boxes), lambda: boxes)

    # normalize image
    norm = tf.keras.applications.resnet.preprocess_input(tf.cast(rgb, tf.float32))

    # expand dimensions
    norm = tf.expand_dims(norm, axis=0)

    return norm, boxes, image_name


def extract_boxes_n_scores(norm: tf.Tensor, backbone: tf.keras.Model, rpnetwork: tf.keras.Model, anchors: tf.Tensor,
                           anchor_px: tf.Tensor, field: int, initial_prediction_threshold: float = 0.3):
    """
    extracts rpn_boxes and corresponding objectiveness scores with the faster r-cnn network


    - N: number of predictions
    - d: feature dimensions

    Parameters
    ----------
    norm: tensor (float32)
        normalized image
    backbone: tf.keras.Model
        backbone network usually resnet-50
    rpnetwork: tf.keras.Model
        region proposal network
    anchors: tf.Tensor
        shape: M x 4
        anchors for the region proposal network
    anchor_px: tf.Tensor
        shape: K
        anchor width and height ratios
    field: int
        field size
    initial_prediction_threshold: float
        objectiveness threshold for predictions to be passed to the network
    Returns
    -------
    features: tensor (float32)
        shape: N x d
        features extracted from the image for each box
    rpn_boxes: tensor (float32)
        shape: N x 4
        each prediction in box form
    scores: tensor (float32)
        shape: N x 1
        objectiveness score for each prediction
    """
    # extract features and rpn boxes from image
    features = backbone(norm, training=False)
    outputs = rpnetwork(features, training=False)

    rpn_reg = map_outputs(outputs[1], anchors, anchor_px, field)
    rpn_boxes = unparameterize(rpn_reg, anchors)

    # get objectiveness scores
    rpn_obj = tf.nn.softmax(map_outputs(outputs[0], anchors, anchor_px, field))
    scores = rpn_obj[:, 1] / (tf.reduce_sum(rpn_obj, axis=1))

    # filter out negative predictions
    condition = tf.where(tf.greater(scores, initial_prediction_threshold))
    rpn_boxes = tf.gather_nd(rpn_boxes, condition)
    scores = tf.expand_dims(tf.gather_nd(scores, condition), axis=1)

    return features, rpn_boxes, scores



"""
This package contains functions for data augmentation
"""

# make functions available at the package level using shadow imports
from .augmentation import crop
from .augmentation import flip
from .augmentation import jitter
from .augmentation import shrink

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'crop',
    'flip',
    'jitter',
    'shrink',
)



import tensorflow as tf


@tf.function
def roialign(features, boxes, field, pool=2, tiles=3):
    """Performs roialign on a collection of regressed bounding boxes.
    
    Given an array of N boxes, this first calculates the pixel coordinates
    where features should be interpolated within each regressed bounding box,
    then performs this interpolation and pools the results to generate a
    single feature vector for each regressed box. Each regressed box is
    subdivided into a 'tiles' x 'tiles' array, with each tile containing
    'pool' x 'pool' interpolation points. Pooling of interpolated features
    is performed at the tile level, and the resulting pooled features are
    concatenated to form a single feature vector for objectness / class
    classification.
        
    Parameters
    ----------
    features: tensor
        The three dimensional feature map tensor produced by the backbone network.
    boxes: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a regressed box and its width and height in that order.
    field: float32
        Field size of the backbone network in pixels.
    pool: int32
        pool^2 is the number of locations to interpolate features at within 
        each tile.
    tiles: int32
        tile^2 is the number of tiles that each regressed bounding box is divided
        into.
        
    Returns
    -------
    interpolated: tensor
        An N x pool * tiles x pool * tiles x features array containing the 
        interpolated features in a 2D layout grouped by bounding box. These can
        be max-pooled (pool, pool) along the spatial dimensions (dimensions 2 
        and 3) to produce an N x tile x tile x features array.
    """
    
    #generate pixel coordinates of box sampling locations
    offsets = _roialign_coords(boxes, pool*tiles)

    #convert from pixel coordinates to feature map / receptive field coordinates
    #the top-left feature is centered at (field/2, field/2) pixels, so an offset needs
    #to be applied when comparing features which are located in the center of each
    #receptive fields with boxes that are defined by their upper left corner in the
    #image.
    offsets = (offsets - field / 2) / tf.cast(field, tf.float32)
    
    #bilinear interpolation
    interpolated = _bilinear(features, offsets[:,0], offsets[:,1])

    #reshape so that first dimension is the box id, second/third are the spatial dimensions
    #of the subdivided box, and the last dimension is the feature
    interpolated = tf.reshape(interpolated, [tf.shape(boxes)[0],
                                             tf.cast(pool*tiles, tf.int32),
                                             tf.cast(pool*tiles, tf.int32),
                                             tf.shape(features)[-1]])

    return interpolated


@tf.function
def _roialign_coords(boxes, n_points):
    """Generates arguments for bilinear interpolation of feature maps
    for a single bounding box.
    
    Used for performing roialign on regressed bounding boxes from the
    region proposal network output. This generates the coordinates
    needed to perform bilinear interpolation across the region proposal
    network feature maps at the internal grid points in the regressed
    box.
        
    Parameters
    ----------
    boxes: tensor
        A 2d tensor where each row represents a box, containing 4 elements 
        describing the x,y location of the upper left corner of a regressed 
        box and its width and height in that order.
    n_points: float
        The number of points to interpolate at across the width / height of 
        the regressed box.
        
    Returns
    -------
    sample: tensor
        The interpolation coordinate arguments in pixel units for use with 
        tfa.image.interpolate_bilinear. These are an n_points^2 x 2 array
        of x, y coordinates for interpolation that are specific to the position
        and shape of the input 'box'.
    """

    #generate a sequence of the indices of sample locations
    indices = tf.range(0.0, n_points)

    #calculate horizontal, vertical spacings between samples (in pixels)
    delta_width = boxes[:,2] / tf.cast(tf.size(indices), tf.float32)
    delta_height = boxes[:,3] / tf.cast(tf.size(indices), tf.float32)

    #calculate vertical and horizontal positions of sample locations
    x_offset = tf.expand_dims(delta_width / 2.0, axis=1) + \
        tf.tensordot(delta_width, indices, axes=0)
    y_offset = tf.expand_dims(delta_height / 2.0, axis=1) + \
        tf.tensordot(delta_height, indices, axes=0)

    #generate pairs of all possible x, y coordinates for each box
    x_offset = tf.tile(x_offset, [1, n_points])
    y_offset = tf.repeat(y_offset, n_points*tf.ones(n_points, tf.int32), axis=1)

    #add box corner x,y coordinates to offsets
    x_offset = x_offset + tf.expand_dims(boxes[:,0], axis=1)
    y_offset = y_offset + tf.expand_dims(boxes[:,1], axis=1)

    #reshape offsets to a 2-D array with x,y pairs in rows
    sample = tf.reshape(tf.stack([x_offset, y_offset], axis=2),
                        [n_points * n_points * tf.shape(boxes)[0], 2])

    return sample


@tf.function
def _bilinear(features, x, y):
    """
    Performs bilinear interpolation of a 3d tensor along first two
    dimensions. Used for calculations of roialign. Replacement for 
    tensorflow addons bilinear interpolation.
    
    Parameters
    ----------
    features: tensor (float32)
        A three-dimensional tensor (M, N, K) where the third dimension are
        features / channels.    
    x: tensor (float32)
        Horizontal coordinates where linear interpolation is desired.
    y: tensor (float32)
        Vertical coordinates where linear interpolation is desired.    
        
    Outputs
    -------
    interpolated: tensor (float32)
        An (d, 1, K) tensor where d is the number of points to interpolate.
    """    
    
    #calculate top/bottom, left/right reference positions
    lower_y, upper_y = _linear_indices(y, tf.shape(features)[-3])
    lower_x, upper_x = _linear_indices(x, tf.shape(features)[-2])

    #feature values at horizontal reference locations
    fly_lx, fly_ux = _linear_f(features, lower_y, lower_x, upper_x, axis=1)
    fuy_lx, fuy_ux = _linear_f(features, upper_y, lower_x, upper_x, axis=1)

    #horizontal linear interpolation
    fly = _linear_interp(fly_lx, fly_ux, x, lower_x)
    fuy = _linear_interp(fuy_lx, fuy_ux, x, lower_x)

    #vertical linear interpolation
    interpolated = _linear_interp(fly, fuy, y, lower_y)

    #reshape output to return 3d tensor
    interpolated = tf.reshape(interpolated,
                               [tf.size(x), 1, tf.shape(features)[-1]])
    
    return interpolated


@tf.function
def _linear_indices(x, size):
    """
    Calculates locations of upper and lower reference points used for
    interpolation.
    
    Parameters
    ----------
    x: tensor (float32)
        Coordinates where linear interpolation is desired (either
        x or y values). 
    size: int
        Size of the tensor to interpolate along that dimension.
        
    Outputs
    -------
    lower: tensor (int32)
        Index for left/top location to use for interpolation.
    upper: tensor (int32)
        Index for right/bottom location to use for interpolation.
    """
    
    #calculate lower and upper indices for interpolation
    lower = tf.cast(tf.minimum(tf.maximum(0.0, tf.math.floor(x)), 
                               tf.cast(size-1, tf.float32)), tf.int32)
    upper = tf.cast(tf.minimum(lower+1, size-1), tf.int32)
    
    return lower, upper


@tf.function
def _linear_f(features, fixed, lower, upper, axis):
    """
    Extracts feature values at left/right or top/bottom reference
    points.
    
    Parameters
    ----------
    features: tensor (float32)
        An M x N x K tensor containing features along dimension 2.
    fixed: tensor (float32)
        A d-length tensor containing the fixed coordinates (x or y) to be used 
        in the interpolation. Bilinear interpolation is performed along each axis, 
        holding the other fixed.
    lower: tensor (float32)
        A d-length tensor containing the coordinates of the top/left reference 
        points.
    upper: tensor (float32)
        A d-length tensor containing the coordinates of the bottom/right 
        reference points.
    axis: int
        The axis of features that the interpolation will be performed against.
        Either 0 or 1.
    
    Outputs
    -------
    f_lower: tensor (float32)
        A d x K tensor containing the feature values of top/left reference points
        in rows.
    f_upper: tensor (float32)
        A d x K tensor containing the feature values of bottom/right reference 
        points in rows.        
    """
    
    #build tensors for calling gather_nd - assume axis=1 ('fixed' is y)
    lower_indices = tf.stack([tf.cast(fixed, tf.int32), lower], axis=1)
    upper_indices =  tf.stack([tf.cast(fixed, tf.int32), upper], axis=1)

    #if vertical interpolation (axis=0), switch indices for gather_nd
    lower_indices = tf.gather(lower_indices, [axis, 1-axis], axis=1)
    upper_indices = tf.gather(upper_indices, [axis, 1-axis], axis=1)

    #transform to linear indices to use gather instead of gather_nd and avoid OOM
    lower_linear = lower_indices[:,1] * tf.shape(features)[-2] + lower_indices[:,0]
    upper_linear = upper_indices[:,1] * tf.shape(features)[-2] + upper_indices[:,0]

    #create a reshaped view of 'features' for linear indexing
    reshaped = tf.reshape(features, [-1, tf.shape(features)[-1]])

    #calculate slopes
    f_lower = tf.gather(reshaped, lower_linear)
    f_upper = tf.gather(reshaped, upper_linear)

    return f_lower, f_upper


@tf.function
def _linear_interp(f_lower, f_upper, x, lower):
    """
    One-dimensional interpolation given the features values at reference points.
    
    Parameters
    ----------
    f_lower: tensor (float32)
        A d x K array containing feature values at d left/top reference points.
    f_upper: tensor (float32)
        A d x K array containing feature values at d right/bottom reference points.
    x: tensor (float32)
        A d-length array containing the horizontal/vertical coordinates to interpolate at.
    lower: tensor (float32)
        A d-length array containing the coordinates of the left/top reference points.
    
    Outputs
    -------
    interpolated: tensor (float32)
        A d x K array containing interpolated feature values at x.
    """
    
    #calculate offsets between x and lower
    dx = tf.maximum(0.0, x-tf.cast(lower, tf.float32))
    
    #calculate output
    interpolated = f_lower + tf.multiply(tf.expand_dims(dx, axis=-1), 
                                         f_upper-f_lower)  
    
    return interpolated



from .iou import greedy_iou_mapping, iou
import tensorflow as tf


class AveragePrecision(tf.keras.metrics.Metric):
    """
    This metric measures the average precision at a specific intersection-over-union
    threshold. The state consists of tensors describing true positive, false positive,
    and false negative detections at various objectness thresholds.
    
    Attributes
    ----------
    iou_thresh: float32
        Threshold for intersection-over-union to determine when a predicted box 
        has detected a ground truth box. Default value 0.5.
    delta: float32
        Increment for thresholds applied to objectness scores for predicted boxes.
        Range is (0, 1). Default value 0.1.
    """
    
    def __init__(self, iou_thresh=0.5, delta=0.1, name='average_precision', **kwargs):
        """
        The constructor accepts the 'iou_thresh' and 'delta' attributes.
        """
        
        super(AveragePrecision, self).__init__(name=name, **kwargs)

        #set iou threshold and objectness increment
        self.iou_thresh = iou_thresh
        self.delta = delta

        #calculate number of thresholds
        n = tf.size(tf.range(0, 1, delta))

        #set states to zero
        self.tp = self.add_weight(name='tp', shape=n, initializer='zeros')
        self.fp = self.add_weight(name='fp', shape=n, initializer='zeros')
        self.fn = self.add_weight(name='fn', shape=n, initializer='zeros')
    

    def update_state(self, y_true, y_pred, sample_weight=None):
        """
        Updates the true positive, false positive, and false negative states
        based on predictions and ground truth for a single image.
        
        Parameters
        ----------
        y_true: tensor (float32)
            An M x 4 tensor of ground truth boxes where each row contains
            the x, y coordinates of the upper-left box corner and the box 
            width and height.
        y_pred: tensor (float32)
            An N x 6 tensor containing the objectness scores and regressions
            for N predictions. The first two columns contain the objectness 
            score outputs. The last four columns contain the x, y coordinates
            of the upper-left box corners and the box widths and heights.            
        """
    
        #split predictions into objectness, regression
        obj = tf.gather(y_pred, [0, 1], axis=1)
        reg = tf.gather(y_pred, tf.constant([2, 3, 4, 5], tf.int32), axis=1)

        #generate objectness threshold sequence
        taus = tf.range(0, 1, self.delta)
    
        def threshold(obj, reg, boxes, tau, min_iou):

            #call objects from rpn
            positive = tf.greater_equal(obj[:,1], tau)
            obj = tf.boolean_mask(obj, positive, axis=0)
            reg = tf.boolean_mask(reg, positive, axis=0)

            #greedy mapping regressions to ground-truth boxes
            ious = iou(reg, boxes)
            tp, fp, fn, tp_list, fp_list, fn_list = greedy_iou_mapping(ious, min_iou)        

            return tf.stack([tp, fp, fn], axis=0)
    
        #calculate precision-recall values
        packed = tf.map_fn(lambda x: threshold(obj, reg, y_true, x, self.iou_thresh),
                           taus, parallel_iterations=10, fn_output_signature=tf.int32)

        #unpack
        tp, fp, fn = tf.unstack(packed, axis=1)  

        #update
        self.tp.assign_add(tf.cast(tp, tf.float32))
        self.fp.assign_add(tf.cast(fp, tf.float32))
        self.fn.assign_add(tf.cast(fn, tf.float32))
    

    def result(self):
        """
        Returns the average precision calculated over self.delta width steps for objectness
        score.
        """
        
        #calculate precision
        precision = self.tp / (self.tp + self.fp)

        #add 1 to end
        precision = tf.concat([precision, [1.0]], axis=0)

        #average precision
        ap = tf.reduce_mean(precision)

        return ap


    def reset_state(self):
        """
        Resets the true positive, false positive, and false negative states to zero.
        """
        
        self.tp.assign(tf.zeros(self.tp.shape))
        self.fp.assign(tf.zeros(self.fp.shape))
        self.fn.assign(tf.zeros(self.fn.shape))



import tensorflow as tf
from scipy.optimize import linear_sum_assignment

from histomics_detect.metrics.iou import iou


def cluster_assignment(boxes: tf.Tensor, rpn_positive: tf.Tensor, min_threshold: float = 0.0,
                       apply_threshold: bool = False) -> tf.Tensor:
    """
    calculates the cluster assignment of the predictions to the ground truth boxes
    a cluster is a group of predictions all of which are closest to the same ground truth box

    !!Assumption: each ground truth box has at least one corresponding prediction (e.g. ~ N >> G)
                  this assumption does not need to hold for this function to run, but it needs to hold
                  for the output to make sense

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_positive: tensor (float32)
        predictions
        shape: N x 4
    min_threshold: float
        if box has no ground truth with an iou higher than 'min_threshold' this box is considered an outlier
        and is not assigned to a cluster
    apply_threshold: bool
        set assignment of boxes with overlap less than threshold to -1

    Returns
    -------
    clusters: tensor (int32)
        a list with one element per prediction (rpn_positive)
        that element is the index of the closest ground truth box

    """
    ious = iou(rpn_positive, boxes)

    def assign_single_prediction(i) -> tf.int32:
        assignment = tf.cast(tf.argmax(ious[i]), tf.int32)
        if apply_threshold:
            assignment = tf.cond(ious[i, assignment] > min_threshold, lambda: assignment,
                                 lambda: tf.constant(-1, dtype=tf.int32))
        return assignment

    clusters = tf.map_fn(assign_single_prediction, tf.range(0, tf.shape(rpn_positive)[0]))
    return tf.cast(clusters, tf.int32)


@tf.function
def tf_linear_sum_assignment(boxes, rpn_boxes):
    ious = iou(boxes, rpn_boxes)

    out = tf.numpy_function(linear_sum_assignment, [ious, tf.constant(True)], [tf.int64, tf.int64])
    row_ind, col_ind = out[0], out[1]

    return tf.expand_dims(tf.cast(col_ind, tf.int32), axis=1)





import tensorflow as tf


@tf.function
def parameterize(positive, boxes):
    """Transforms boxes that are matched to an anchor to a parameterized format
    used by the regression loss.
        
    Parameters
    ----------
    positive: tensor (float32)
        M x 5 tensor of anchors matched to ground truth boxes. Each row contains
        the x,y upper left corner, width, height, and matched box index for one
        anchor.
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
        
    Returns
    -------
    parameterized: tensor (float32)
        N x 4 tensor of anchor-matched boxes in a parameterized format.
    """

    #gather boxes matched to each anchor
    matched = tf.gather(boxes, tf.cast(positive[:,4], tf.int32), axis=0)

    #calculate parameterization using matched anchors
    tx = tf.divide(matched[:,0] - positive[:,0], positive[:,2])
    ty = tf.divide(matched[:,1] - positive[:,1], positive[:,3])
    tw = tf.math.log(tf.divide(matched[:,2], positive[:,2]))
    th = tf.math.log(tf.divide(matched[:,3], positive[:,3]))

    #stack results
    parameterized = tf.stack([tx, ty, tw, th], axis=1)

    return parameterized


@tf.function
def unparameterize(parameterized, positive):
    """Transforms parameterized boxes to an un-parameterized [x,y,w,h] format.
    
    Given the matched and row-aligned tensors containing parameterized boxes and
    corresponding positive anchors, reverts the parameterization so that boxes
    can be used in IoU calculations or visualization.
        
    Parameters
    ----------
    parameterized: tensor (float32)
        N x 4 tensor of anchor-matched boxes in a parameterized format.
    positive: tensor (float32)
        N x 5 tensor of anchors matched to ground truth boxes. Each row contains
        the x,y upper left corner, width, height, and matched box index for one
        anchor.
        
    Returns
    -------
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    """
    
    #convert from parameterized representation to a box representation

    #transform predictions back to box format
    x = tf.multiply(parameterized[:,0], positive[:,2]) + positive[:,0]
    y = tf.multiply(parameterized[:,1], positive[:,3]) + positive[:,1]
    w = tf.multiply(tf.exp(parameterized[:,2]), positive[:,2])
    h = tf.multiply(tf.exp(parameterized[:,3]), positive[:,3])

    #stack
    boxes = tf.stack([x, y, w, h], axis=1)

    return boxes
  

@tf.function
def tf_box_transform(boxes):
    """Transform bounding boxes to TF format.
    
    Transforms the [x,y,w,h] format for bounding boxes used throughout this package
    to the [x,y,x+w-1,x+h-1] expected by TensorFlow.
    
    Parameters
    ----------
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
        
    Returns
    -------
    transformed: tensor (float32)
        N x 4 tensor where each row contains the x,y locations of the upper left
        and lower right corners in that order.
    """

    #unstack box columns
    x, y, w, h = _unstack_box_array(boxes)

    #join transformed columns
    transformed = tf.stack([x, y, tf.add(x, w-1), tf.add(y, h-1)], axis=1)

    return transformed
    

@tf.function
def clip_boxes(boxes, width, height):
    """Clips boxes that extend beyond a region of interest boundary.
    
    Transforms each box so that the minimum x/y coordinates are zero and the maximum
    coordinates are width/height.
    
    Parameters
    ----------
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    width: float32
        Width of region of interest to clip boxes to.
    height: float32
        Height of region of interest to clip boxes to.
        
    Returns
    -------
    clipped: tensor (float32)
        Same as input but with boxes clipped to region of interest.
    """    
    
    
    #clips bounding boxes in [x,y,w,h] format to image dimensions

    # unstack box columns
    x, y, w, h = _unstack_box_array(boxes)

    #clip left corner
    x = tf.maximum(x, 0.0)
    y = tf.maximum(y, 0.0)

    #clip edge lengths
    w = tf.subtract(tf.minimum(tf.cast(width, tf.float32), tf.add(x,w)), x)
    h = tf.subtract(tf.minimum(tf.cast(height, tf.float32), tf.add(y,h)), y)

    #join clipped columns
    clipped = tf.stack([x,y,w,h], axis=1)

    return clipped


@tf.function
def _unstack_box_array(boxes):
    """Unstacks the x,y,w,h bounding box parameters from the stacked input.
    
    Parameters
    ----------
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
        
    Returns
    -------
    x: tensor (float32)
        N-length tensor of x coordinates of upper left corner of bounding boxes.
    y: tensor (float32)
        N-length tensor of y coordinates of upper left corner of bounding boxes.    
    w: tensor (float32)
        N-length tensor of bounding boxe widths.
    h: tensor (float32)
        N-length tensor of bounding boxe heights.
    """    
    
    
    #unstacks columns from bounding box array. use gather instead of unstack, 
    #because input could have 5 columns (arrays containing 'positive' anchors that
    #have been matched to bounding boxes can contain a fifth column that is the
    #index of the matched bounding box. 
  
    #gather first four columns
    x = tf.gather(boxes, 0, axis=1)
    y = tf.gather(boxes, 1, axis=1)
    w = tf.gather(boxes, 2, axis=1)
    h = tf.gather(boxes, 3, axis=1)

    return x, y, w, h


@tf.function
def filter_edge_boxes(boxes, width: float, height: float, margin: float = 32.0,
                      centroids=tf.constant(True, tf.bool)):
    """
    Filters out boxes that cross the margin of the image boundary.

    Boxes are retained if all the points are within the image boundary inset by the 
    margin parameter. Ex. given a box with top left and bottom right corners [[2, 2], 
    [9, 9] for a image with size 10, 10 this box is not filtered if margin <= 1.

    Parameters
    ----------
    boxes: tensor (float32)
        M x 4 tensor of bounding boxes where each row contains the x,y location of 
        the upper left corner of a ground truth box and its width and height in that 
        order.
    width: float
        Width of the image that boxes originate from.
    height: float
        Height of the image that boxes originate from. 
    margin: float
        Offset from the border of the image border where boxes that overlap are 
        removed. Default value 5.
    centroid: boolean.
        If True, the filtering removes boxes whose centroids fall within the margin. 
        If False, any box intersecting the margin will be filtered. Default value 
        True.
        
    Returns
    -------
    filtered_boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
        All boxes that have at least one point in the margin have been removed.
    mask: tensor (bool)
        M length tensor where true indicates box was retained.
    """

    # cast inputs to float32
    margin = tf.cast(margin, tf.float32)
    width = tf.cast(width, tf.float32)
    height = tf.cast(height, tf.float32)

    # unstack box columns
    x, y, w, h = _unstack_box_array(boxes)

    # condition for boxes to be kept
    def box(x, y, w, h):
        min_cond = tf.logical_and(x >= margin, y >= margin)
        max_cond = tf.logical_and(x+w <= (width-margin), y+h <= (height-margin))
        return tf.logical_and(min_cond, max_cond)        
    
    # condition for centroids to be kept
    def centroid(x, y, w, h):
        min_cond = tf.logical_and(x+w/2 >= margin, y+h/2 >= margin)
        max_cond = tf.logical_and(x+w/2 <= (width-margin), y+h/2 <= (height-margin))
        return tf.logical_and(min_cond, max_cond) 
    
    # generate mask
    mask = tf.cond(centroids, lambda: centroid(x,y,w,h), lambda: box(x,y,w,h))

    # stack columns and collect boxes that fulfill the condition
    filtered_boxes = tf.gather_nd(tf.stack([x, y, w, h], axis=1), tf.where(mask))
    
    return filtered_boxes, mask



import tensorflow as tf
from typing import Tuple, Union
from scipy.optimize import linear_sum_assignment

from histomics_detect.metrics.iou import iou


@tf.function
def tf_linear_sum_assignment(boxes, rpn_boxes):
    """
    tensorflow wrapper for linear sum assignment function from scipy.optimice

    Parameters
    ----------
    boxes: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a box and its width and height in that order. Typically the ground truth boxes
    rpn_boxes: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a box and its width and height in that order. Typically the
        predictions.

    Returns
    -------
    true_rpn_box_indexes: tensor (int32)
        returns the indexes of the rpn_boxes assigned to the corresponding gt box

    """
    ious = iou(boxes, rpn_boxes)
    out = tf.numpy_function(linear_sum_assignment, [ious, tf.constant(True)], [tf.int64, tf.int64])
    row_ind, col_ind = out[0], out[1]

    return tf.expand_dims(tf.cast(col_ind, tf.int32), axis=1)



import tensorflow as tf

def iou(source, target):
    """Calculates intersection over union (IoU) and intersection areas for two sets
    of objects with box representations.
    
    This uses simple arithmetic and outer products to calculate the IoU and intersections
    between all pairs without looping.
        
    Parameters
    ----------
    source: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a box and its width and height in that order. Typically the
        predictions.
    target: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a box and its width and height in that order. Typically the
        ground truth.
        
    Returns
    -------
    iou: tensor (float32)
        M x N tensor containing IoU values between source and target boxes.
    intersection: tensor (float32)
        M x N tensor containing area overlaps between source and target boxes
        in pixels.
    """
    
    #split into corners and sizes
    xs, ys, ws, hs = tf.split(source, 4, axis=1)
    xt, yt, wt, ht = tf.split(target, 4, axis=1)

    #overlap in dimensions
    left = tf.maximum(xs, tf.transpose(xt))
    top = tf.maximum(ys, tf.transpose(yt))
    right = tf.minimum(xs+ws, tf.transpose(xt+wt))
    bottom = tf.minimum(ys+hs, tf.transpose(yt+ht))

    horizontal = tf.minimum(xs+ws, tf.transpose(xt+wt)) - \
        tf.maximum(xs, tf.transpose(xt))
    vertical = tf.minimum(ys+hs, tf.transpose(yt+ht)) - \
        tf.maximum(ys, tf.transpose(yt))

    #calculate intersection
    intersection = tf.maximum(0.0, horizontal) * tf.maximum(0.0, vertical)

    #calculate iou
    iou = intersection / (ws*hs + tf.transpose(wt*ht) - intersection)

    return iou


def _greedy_iou_mapping_iter(i, ious, source_mask, target_mask, matches):
    """Performs one iteration of greedy IoU mapping.
    
    This is the loop body of the greedy IoU mapping algorithm. This identifies the 
    best match having the highest IoU and removes the corresponding prediction and 
    ground truth element from future consideration in matching.
    
    Parameters
    ----------
    i: int32
        Iteration number in mapping. Used for writing to output TensorArray.
    ious: tensor (float32)
        M x N tensor of IoU values used to generate mapping. Regression predictions
        are in rows and ground truth elements are in columns. This array is masked
        to remove previous matches when identifying the highest IoU match.
    source_mask: tensor (bool)
        1D M-length tensor where unmatched predictions are represented by 'True'.
    target_mask: tensor (bool)
        1D M-length tensor where unmatched ground truth elements are represented 
        by 'True'.    
    matches: tensor (float32)
        2D tensor where each row represents a match, containing the indices of
        the matched prediction and ground truth element in that order.
        
    Returns
    -------
    i: int32
        Loop iteration counter.
    ious: tensor (float32)
        Same as input but updated with current iteration match.
    source_mask: tensor (bool)
        Same as input but updated with current iteration match.
    target_mask: tensor (bool)
        Same as input but updated with current iteration match.
    matches: tensor (float32)
        Same as input but updated with current iteration match.
    """

    #mask targets and get best match for each source
    maxima = tf.reduce_max(tf.boolean_mask(ious, target_mask, axis=1), axis=1)
    target_indices = tf.argmax(tf.boolean_mask(ious, target_mask, axis=1), axis=1)

    #mask sources that were already matched
    maxima = tf.boolean_mask(maxima, source_mask)
    target_indices = tf.boolean_mask(target_indices, source_mask)

    #get source and target indices
    max = tf.reduce_max(maxima)
    source_index = tf.argmax(maxima)
    target_index = tf.gather(target_indices, source_index)

    #correct for masked sources and targets
    source_index = tf.gather(tf.where(source_mask), source_index)
    target_index = tf.gather(tf.where(target_mask), target_index)

    #update masks
    source_mask = tf.tensor_scatter_nd_update(source_mask, [source_index],
                                              [tf.constant(False)])
    target_mask = tf.tensor_scatter_nd_update(target_mask, [target_index],
                                              [tf.constant(False)])
  
    #write (source, target) to TensorArray
    matches = matches.write(i, tf.concat([tf.cast(source_index, tf.float32),
                                          tf.cast(target_index, tf.float32),
                                          [max]], axis=0))
  
    #update index
    i = i + 1

    return i, ious, source_mask, target_mask, matches


def greedy_iou_mapping(ious, min_iou):
    """Calculates greedy IoU mapping between predictions and ground truth.
    
    Uses intersection-over-union scores to compute a greedy mapping between
    ground truth and predicted objects. Greedy mapping can produce suboptimal
    results compared to the Kuhn–Munkres algorithm since matching is greedy.
    
    Parameters
    ----------
    ious: tensor (float32)
        M x N tensor of IoU values used to generate mapping. Regression predictions
        are in rows and ground truth elements are in columns. This array is masked
        to remove previous matches when identifying the highest IoU match.
    min_iou: float32
        Minimum IoU threshold for defining a match between a regression
        prediction and a ground truth box.
        
    Returns
    -------
    precision: float32
        Precision of IoU mapping.
    recall: float32
        Recall of IoU mapping.
    tp: int32
        True positive count of IoU mapping.
    fp: int32
        False positive count of IoU mapping.
    fn: int32
        False negative count of IoU mapping.
    tp_list: int32
        Two-dimensional tensor containing indices of true positive predictions
        in first column, and corresponding matching ground truth indices in second
        column.
    fp_list: int32
        One-dimensional tensor containing indices of false positive predictions.
    fn_list: int32
        One-dimensional tensor containing indices of false negative ground truth.
    """

    #initialize masks
    source_mask = tf.ones(tf.shape(ious)[0], tf.bool)
    target_mask = tf.ones(tf.shape(ious)[1], tf.bool)

    #define loop counter, condition, store for output
    i = tf.constant(0)
    matches = tf.TensorArray(tf.float32, size=tf.shape(ious)[0], dynamic_size=False)
    condition = lambda i, a, b, c, d: tf.less(i, tf.minimum(tf.shape(ious)[0],
                                                            tf.shape(ious)[1]))

    #loop to perform greedy mapping
    _, _, _, _, matches = tf.while_loop(condition, _greedy_iou_mapping_iter,
                                        [i, ious, source_mask, target_mask, matches],
                                        parallel_iterations=10)

    #stack outputs
    matches = matches.stack()

    #discard matches that do not meet min_iou
    matches = tf.boolean_mask(matches,
                              tf.greater_equal(matches[:,2], min_iou), axis=0)

    #calculate TP, FP, FN, precision, recall
    tp = tf.shape(matches)[0]
    fp = tf.shape(ious)[0] - tf.shape(matches)[0]
    fn = tf.shape(ious)[1] - tf.shape(matches)[0]

    #generate lists of indexes for TP, FP, FN
    tp_list = tf.cast(matches[:,0:2], tf.int32)
    fp_list = tf.sets.difference([tf.range(tf.shape(ious)[0], dtype=tf.int32)],
                                 [tf.cast(matches[:,0], dtype=tf.int32)]).values
    fn_list = tf.sets.difference([tf.range(tf.shape(ious)[1], dtype=tf.int32)],
                                 [tf.cast(matches[:,1], dtype=tf.int32)]).values

    return tp, fp, fn, tp_list, fp_list, fn_list



import os
from PIL import Image
import tensorflow as tf


def dataset(path, png_parser, csv_parser, size, cases=None):
    """Generates a tf.data.Dataset object containing matched region
    of interest .pngs and bounding box .csv files.
    
    This function accepts parser function handles that are used to parse
    and match png and csv filenames. Each parser is written to return 
    case and unique roi identifier strings that are used for matching
    png and csv files and for filtering cases from the match list.
        
    Parameters
    ----------
    path: string
        Path containing png region of interest images and csv files 
        containing corresponding bounding boxes.
    png_parser: function
        Function that accepts a single string containing the filename
        of a png image (without path) and that returns the corresponding
        case and roi name. The case and roi name should uniquely identify
        the roi within the dataset.
    csv_parser: function
        Function that accepts a single string containing the filename
        of a png image (without path) and that returns the corresponding
        case and roi name. The case and roi name should uniquely identify
        the roi within the dataset.
    cases: list of strings
        An optional list of cases used to select rois for inclusion in the 
        dataset. Default value of None will select all cases in 'path'.
        Helpful when cases are grouped by folder
        
    Returns
    -------
    ds: tf.data.Dataset
        A dataset where each element contains an rgb image tensor, an N x 4
        tensor of bounding boxes where each row contains the x,y location 
        of the upper left corner of a ground truth box and its width and
        height in that order, and the png filename.
    """
    
    #get list of csv and png files in path   
    csvs = [f for f in os.listdir(path) if os.path.splitext(f)[1] == '.csv']
    pngs = [f for f in os.listdir(path) if os.path.splitext(f)[1] == '.png']
        
    #extract case, roi strings from filenames
    csv_case_roi = [csv_parser(csv) for csv in csvs]
    png_case_roi = [png_parser(png) for png in pngs]
        
    #form lists of case + roi for matching
    csv_match_string = [csv[0] + csv[1] for csv in csv_case_roi]
    png_match_string = [png[0] + png[1] for png in png_case_roi]
        
    #match
    indexes = [csv_match_string.index(png) if png in csv_match_string else -1 for 
               png in png_match_string]
    
    #form tuples of case, matching png file, csv file
    matches = [(case_roi[0], png, csvs[index]) for (case_roi, png, index) in 
               zip(png_case_roi, pngs, indexes) if index != -1]
        
    #filter on cases
    if cases is not None:
        matches = [match for match in matches if match[0] in cases]
            
    #format outputs
    matches = [(path + match[1], path + match[2]) for match in matches]
    
    #filter on image size
    matches = [(png, csv) for (png, csv) in matches if
               (Image.open(png).size[0] > size) and (Image.open(png).size[1] > size)]
    
    #create tf.data.Dataset object of pairs of csv/png files
    ds = tf.data.Dataset.from_tensor_slices(matches)
    
    #map image and csv read operations to generate (rgb, boxes, pngfile) tuple
    ds = ds.map(lambda x: (read_png(x[0]), read_csv(x[1]), x[0]))
    
    return ds


@tf.function
def read_csv(csv_file):
    """
    Reads a csv file describing bounding boxes using tensorflow operations.
    
    Each file should include a single header row, and each subsequent rows
    describes a single box and contains the fields
    
        x (float) - horizontal upper-left corner of box (pixels)
        y (float) - vertical upper-left corner of box (pixels)
        w (float) - box width (pixels)
        h (float) - box height (pixels)
        label (string) - class label (optional)
        center_x (float - optional) - horizontal center of box (for point annotations)
        center_y (float - optional) - vertical center of box (for point annotations)
        slide_x (float - optional) - horizontal upper-left corner of box in slide
        slide_y (float - optional) - vertical upper-left corner of box in slide
        type (string - optional) - the type of annotation, either 'box' or 'point'
        contained (bool - optional) - if box or point is entirelycontained within roi
        
    Parameters
    ----------
    csv_file: string
        Path and filename of the csv file.
        
    Returns
    -------
    ds: RaggedTensor
        A ragged tensor where the each row contains the x,y location 
        of the upper left corner of a ground truth box and its width and
        height in that order.
    """
    
    #read contents of csv
    contents = tf.io.read_file(csv_file)

    #split into lines
    lines = tf.strings.split(contents, '\n')

    #decode data lines
    lines = tf.io.decode_csv(lines[1:-1],
                             [0.0, 0.0, 0.0, 0.0, '', 0.0, 0.0, 0.0, 0.0, '', ''])
    
    #embed in ragged tensor
    boxes = tf.RaggedTensor.from_tensor(tf.stack((lines[0], lines[1], 
                                                  lines[2], lines[3]), 
                                                 axis=1))

    return boxes


@tf.function
def read_png(png_file):
    """
    Reads a png file using tensorflow operations.
        
    Parameters
    ----------
    png_file: string
        Path to the png file.
        
    Returns
    -------
    rgb: tensor
        Three dimensional rgb image tensor.
    """
    
    #read in png and get size
    rgb = tf.io.decode_png(tf.io.read_file(png_file))
    
    return rgb


@tf.function
def resize(rgb, boxes, factor):
    """Resizes input image and bounding boxes.
    
    This function can be used as a transformation in the input pipeline to 
    resize the images and bounding boxes. This can help to improve performance
    in images with small objects or images where object density is high
    relative to anchor density.
        
    Parameters
    ----------
    rgb: tensor
        The image as a 2D or 3D tensor.
    boxes: tensor (float32)
        N x 4 tensor containing boxes where each row contains the x,y 
        location of the upper left corner of a ground truth box and its width 
        and height in that order.
    factor: float32
        The scalar factor used for resizing. A value of 2.0 will double the
        magnification.
        
    Returns
    -------
    rgb: tensor
        The input image resized by factor.
    boxes: tensor (float32)
        N x 4 tensor containing boxes resized by factor.
    """
    
    #resize image
    rgb = tf.image.resize(rgb, tf.cast([factor * tf.cast(tf.shape(rgb)[0], tf.float32), 
                                        factor * tf.cast(tf.shape(rgb)[1], tf.float32)], 
                                       tf.int32))
    
    #resize boxes
    boxes = factor * boxes
    
    return rgb, boxes



from histomics_detect.anchors.create import create_anchors
from histomics_detect.anchors.filter import filter_anchors
from histomics_detect.anchors.sampling import sample_anchors
from histomics_detect.boxes import parameterize, unparameterize, clip_boxes, tf_box_transform, filter_edge_boxes
from histomics_detect.metrics import iou, greedy_iou_mapping, AveragePrecision, FalsePositiveRate, FalseNegativeRate
from histomics_detect.networks.backbones import pretrained, residual
from histomics_detect.networks.rpns import rpn
from histomics_detect.networks.fast_rcnn import fast_rcnn
from histomics_detect.networks.field_size import field_size
from histomics_detect.roialign.roialign import roialign
import tensorflow as tf


@tf.function
def map_outputs(output, anchors, anchor_px, field):
    """Transforms region-proposal network outputs from 3D tensors to 2D anchor arrays.
    
    The region-proposal network outputs 3D tensors containing either objectness scores 
    or parameterized regressions. Given a set of K anchor sizes, the objectness and
    regression tensors produced by the region-proposal network will have sizes 2*K 4*K
    along their third dimension respectively. This function transforms these 3D tensors
    to 2D tensors where the objectness or regressions for anchors are in rows.
        
    Parameters
    ----------
    output: tensor (float32)
        M x N x D tensor containing objectness or regression outputs from the 
        region-proposal network.
    anchors: tensor (float32)
        M*N*K x 4 tensor of anchor positions. Each row contains the x,y upper left
        corner of the anchor in pixels relative in the image coordinate frame, 
        and the anchor width and height.
    anchor_px: tensor (int32)
        K-length 1-d tensor containing the anchor width hyperparameter values in pixel 
        units.
    field: float32
        Edge length of the receptive field in pixels. This defines the area of the 
        image that corresponds to 1 feature map and the anchor gridding.
        
    Returns
    -------
    mapped: tensor (float32)
        M*N*K x 2 array of objectness scores or M*N*K x 4 tensor of regressions
        where each row represents one anchor.
    """
  
    #get anchor size index for each matching anchor
    index = tf.equal(tf.expand_dims(anchors[:,3], axis=1),
                     tf.expand_dims(tf.cast(anchor_px, tf.float32), axis=0))
    index = tf.cast(tf.where(index)[:,1], tf.int32)

    #use anchor centers to get positions of anchors in rpn output
    px = tf.cast((anchors[:,0]+ anchors[:,2]/2) / field, tf.int32)
    py = tf.cast((anchors[:,1]+ anchors[:,3]/2) / field, tf.int32)

    #add new dimension to split outputs by anchor (batch, y, x, anchor, output)
    reshaped = tf.reshape(output, tf.concat([tf.shape(output)[0:-1],
                                             [tf.size(anchor_px)],
                                             [tf.shape(output)[-1] /
                                              tf.size(anchor_px)]],
                                            axis=0))

    #gather from (batch, y, x, anchor, output) space to 2D array where each row is 
    #an anchor and columns are objectness (2-array) or regression (4-array) scores
    mapped = tf.gather_nd(reshaped,
                          tf.stack([tf.zeros(tf.shape(px), tf.int32),
                                    py, px, index],
                                   axis=1))

    return mapped


def faster_rcnn_config():
    """Generates a default configuration for a faster RCNN model, setting default
    parameters for the backbone, rpn, and fast RCNN sub-networks, as well as 
    training and validation settings.
    
    Returns
    -------
    backbone_args: dict
        Contains name of backbone and backbone parameters. Parameters vary 
        by backbone type. Defaults to a resnet50 backbone with a stride of 1
        and 14 residual blocks.    
    rpn_args: dict
        Describes the region proposal architecture, specifying the number
        of layers, kernel sizes, output dimensions, and activations.
    frcnn_args:
        Describes the fast RCNN architecture, including the number and size
        of dense layers, dense layer activations, and the roialign parameters.
    train_args:
        Describes the shape of the training images (uniform), the max number
        of anchors to sample per image during training, the maximum ratio 
        of negative to positive anchors sampled, and the loss weighting of
        region proposal classifier and regression losses.
    validation_args:
        Describes objectness thresholds, iou thresholds for nms, iou threshold
        to evaluate true positive / false positive / false negative rate, and
        a sequence of iou thresholds to evaluate mean average precision.
    """
    
    #feature network parameters
    backbone_args = {'name': 'resnet50v2',
                     'stride': 1, #stride (pixels) in first backbone convolution
                     'blocks': 14} #number of residual blocks to use in backbone

    #rpn network parameters
    rpn_args = {'kernels': [3], #kernel sizes (receptive fields) for rpn convolutions
                'dimensions': [256], #number of kernels per layer
                'activations': ['relu']} #activation for rpn convolutions

    #fast-rcnn network parameters
    frcnn_args = {'units': [1024, 1024], #number of units in fast-rcnn dense layers
                  'activations': ['relu', 'relu'], #activations for each dense layer
                  'pool': 2, #number of tiles to pool during roialign
                  'tiles': 3} #number of tiles to split regressed boxes into during roialign

    #training parameters
    train_args = {'train_shape': (224, 224, 3), #shape of training instances
                  'max_anchors': 256, #maximum number of negative anchors to sample per epoch
                  'np_ratio': 1.0, #largest ratio of negative : positive anchors per batch
                  'lmbda': 10.0, #weighting factor for region-proposal network regression loss
                  'hard_fraction': 0.0} #fraction of sampled negative anchors that are hard

    #validation parameters
    validation_args = {'tau': 0.5, #objectness threshold used to classify anchors at inference
                       'nms_iou': 0.3, #min nms threshold used to filter overlapping objects
                       'tpr_iou': 0.5, #single iou threshold used to calculate tpr/fpr/fnr
                       'margin': 32, #margin of image edge to exclude from validation (pixels)
                       'ap_ious': [0.25, 0.5, 0.75], #ious thresholds to use in evaluating mAP
                       'ap_delta': 0.1} #precision step size for calculating mAP

    return backbone_args, rpn_args, frcnn_args, train_args, validation_args


class FasterRCNN(tf.keras.Model):
    """
    This class implements a faster RCNN model which combines a backbone feature
    extraction network with a region proposal network. RoiAlign is used to 
    interpolate and pool features for the variable-sized proposed boxes to support
    refinement of boxes or downstream operations like classification (future 
    enhancement). Class methods provide different levels of output processing,
    from raw proposed regions and objectness scores, to proposals thresholded by
    objectness score, non-max suppressed, and refined by roialign. Utility 
    functions are available to apply non-max suppression or roialign. Training is
    performed in single image batches.
    
    Attributes
    ----------
    backbone_args: array-like (dict)
        Arguments used to construct model backbone.
    rpn_args: array-like (dict)
        Arguments used to construct region-proposal network.
    frcnn_args: array-like (dict)
        Arguments used to construct fast-rcnn network.
    train_args: array-like (dict)
        Arguments used to specify training behavior.
    validation_args: array-like (dict)
        Arguments used to specify validation behavior.
    backbone: tf.keras.Model
        A fully-convolutional backbone model that produces an M x N x D feature map.
    rpnetwork: tf.keras.Model
        A convolutional model that produces objectness and parameterized regressions.
    fastrcnn: tf.keras.Model
        A fully-connected model that operates on roialigned outputs to produce
        refined regressions. Future enhancements will extend this network to perform
        classification.
    anchor_px: tensor (int32)
        One dimensional tensor of anchor sizes.
    field: float (integer-valued)
        The field size in pixels of the backbone network.
    pool: int32
        RoiAlign parameter. pool^2 is the number of locations to interpolate 
        features at within each tile. Default value 2.
    tiles: int32
        RoiAlign parameter. tile^2 is the number of tiles that each regressed 
        bounding box is divided into. Default value 3.
    lmbda: float32
        Loss weight for balancing objectness and regression losses of region
        proposal network. Default value 10.0.
    max_anchors: int32
        Maximum number of total anchors to sample. Default value 256.
    np_ratio: float32
        Will sample at most negative : positive ratio anchors. Default value 2.0.
    tau: float32
        Threshold in range [0,1] used to select region proposals based on 
        region proposal network objectness scores. Default value 0.5.
    nms_iou: float32
        Intersection over union threshold used to remove redundant proposals
        during non-max suppression. Range is (0, 1]. Default value 0.3.
    tpr_iou: float32
        Minimum intersection over union threshold between a proposal and ground
        truth object to call that proposal a detection. Used in calculating
        test performance statistics.
    margin: int32
        The margin value is used to clear ground truth objects and proposals
        from the edges of test images. All objects intersecting partially or
        wholly with this margin will be removed. Default value 32 pixels.
    objectness_metrics: list (tf.keras.Metric)
        A list of tf.keras.Metric objects that can assess the binary
        classification performance of the objectness score from the region
        proposal network. Defaults to precision-recall auc, true positive rate,
        false negatives, and false positives.
    regression_metrics: list (tf.keras.Metric)
        A list of tf.keras.Metric objects that can assess the regression
        performance region proposal network or RoiAlign refined regressions.
        Defaults to average precision at 0.25, 0.50, and 0.75 iou thresholds
        with an objectness threshold step size of 0.1.
    anchors: tensor(float32)
        Stored anchor locations for preset training size to reduce calculations 
        during training.
    
    Methods
    -------
    call
        Produces roialigned, non-max suppressed, and objectness thresholded
        predictions given 
    constructor
        Initializes model, constructing all networks and capturing all parameters
        needed for training and validation. Captures all data necessary to 
        produce configs needed for saving as a keras model.
    input_size
        Used to set or reset input image size dimensions. Useful during training
        and tiled inference.
    threshold
        Applies a threshold to objectness scores from the region proposal network
        and returns filtered proposed boxes and objectness scores.
    nms
        Applies non-max suppression to region proposals using objectness scores.
    align
        Applies RoiAlign to region proposals.
    raw
        Generates the full unfiltered outputs from the region proposal network
        inclding proposal regressions and their objectness scores.
    call
        Generates the objectness thresholded, non-max suppressed, and RoiAligned
        proposals.
    """
    
    def __init__(self, backbone_args, rpn_args, frcnn_args, train_args, validation_args, anchor_sizes,
                 **kwargs):
        super(FasterRCNN, self).__init__(**kwargs)

        #capture input configuration parameters
        self.backbone_args = backbone_args
        self.rpn_args = rpn_args
        self.frcnn_args = frcnn_args
        self.train_args = train_args
        self.validation_args = validation_args
        self.anchor_sizes = anchor_sizes
        
        #build backbone, rpn, and terminal network
        backbone, preprocessor = pretrained(backbone_args['name'])
        self.backbone = residual(backbone, preprocessor, backbone_args['blocks'], backbone_args['stride'])
        self.rpnetwork = rpn(self.backbone.output.shape[-1], len(anchor_sizes), **rpn_args)
        self.fastrcnn = fast_rcnn(self.backbone.output.shape[-1], **frcnn_args)        

        #capture field, anchor sizes, loss mixing
        self.field = tf.cast(field_size(self.backbone), tf.float32)
        
        #capture roialign parameters
        self.pool = frcnn_args['pool']
        self.tiles = frcnn_args['tiles']
        
        #convert anchor_size to tensor
        self.anchor_px = tf.constant(anchor_sizes, dtype=tf.int32)
        
        #capture training arguments
        self.lmbda = train_args['lmbda'] #loss mixing weights
        self.max_anchors = train_args['max_anchors'] #max negative anchors to sample
        self.np_ratio = train_args['np_ratio'] #max acceptable ratio of negative : positive anchors
        self.hard_fraction = train_args['hard_fraction'] #fraction of sampled anchors that are hard
        
        #capture validation arguments
        self.tau = validation_args['tau']
        self.nms_iou = validation_args['nms_iou']
        self.tpr_iou = validation_args['tpr_iou']
        self.margin = validation_args['margin']
        
        #generate anchors for training efficiency - works for fixed-size training
        self.anchors = create_anchors(self.anchor_px, self.field, 
                                      train_args['train_shape'][0], 
                                      train_args['train_shape'][1])

        #define metrics
        self.objectness_metrics = [tf.keras.metrics.AUC(curve="PR", name='prauc'),
                                   tf.keras.metrics.Recall(name='tpr'),
                                   FalsePositiveRate(name='fpr'),
                                   FalseNegativeRate(name='fnr')]
        self.regression_metrics = [AveragePrecision(iou_thresh = t, 
                                                    delta=validation_args['ap_delta'],
                                                    name='ap' + str(int(100*t)))
                                   for t in validation_args['ap_ious']]


    def get_config(self):
        return {'backbone_args': self.backbone_args,
                'rpn_args': self.rpn_args,
                'frcnn_args': self.frcnn_args,
                'train_args': self.train_args,
                'validation_args': self.validation_args,
                'anchor_sizes': self.anchor_sizes}
    
    
    @classmethod
    def from_config(cls, config):
        #defining this function is required to support keras restore
        return cls(**config)
    
    
    def _update_objectness_metrics(self, objectness, positive):
        """
        Updates objectness metrics that are based on binary classification performance.
        
        Parameters
        ----------
        objectness: tensor
            N x 2 tensor containing corresponding softmax objectness scores in rows.
            Second column contains score for being an object.
        positive: tensor (bool)
            N length bool tensor indicating which rows contain objects that were
            judged positive based on

        Returns
        -------
        metrics: dict
            Returns a dict of updated metric values keyed by metric names.       
        """
        
        #update metrics values
        for metric in self.objectness_metrics:
            metric.update_state(tf.cast(positive, tf.uint32), objectness[:,1])
        
        return {m.name: m.result() for m in self.metrics}
    
    
    def _update_regression_metrics(self, boxes, predictions):
        """
        Updates objectness metrics that are based on binary classification performance.
        
        Parameters
        ----------
        boxes: tensor (float32)
            M x 4 tensor where each row contains the x,y location of the upper left
            corner of a ground-truth box and its width and height in that order.
        predictions: tensor (float32)
            N x 6 tensor where each row contains the objectness scores and regressed
            boxes for one proposal.

        Returns
        -------
        metrics: dict
            Returns a dict of updated metric values keyed by metric names.       
        """
        
        #update metrics values
        for metric in self.regression_metrics:
            metric.update_state(boxes, predictions)
        
        return {m.name: m.result() for m in self.metrics}
    
        
    def input_size(self, size=[None, None]):
        """
        Sets input size dimensions for the backbone and region proposal networks. 
        By default, the backbone and region proposal network have variable input image '
        sizes. In some cases this may reduce the speed of training or inference. When
        training or performing inference on uniform sized images, this procedure allows
        the input to be set to a specific size. Default parameters reset these networks 
        to variable size inputs.

        Parameters
        ----------
        size: list of integers
            The height and width of the input images for the backbone. Size of input to
            region proposal network is calculated by devault. Default value is 
            [None, None] for a variable input size.
        """
        
        #get current backbone and rpnetwork channels
        backbone_channels = self.backbone.layers[0].input_shape[0][-1]
        rpnetwork_channels = self.rpnetwork.layers[0].input_shape[0][-1]
        
        #calculate rpnetwork input size from field size and input size
        if all([dim is None for dim in size]):
            rpnetwork_size = [None, None, rpnetwork_channels]
        else:
            rpnetwork_size = [tf.cast(tf.math.ceil(size[0]/self.field), tf.int32),
                              tf.cast( tf.math.ceil(size[1]/self.field), tf.int32), 
                              rpnetwork_channels]
        
        #pop input layers
        self.backbone.layers.pop(0)
        self.rpnetwork.layers.pop(0)
                              
        #create new input layers
        backbone_input = tf.keras.Input(shape = size + [backbone_channels])
        rpnetwork_input = tf.keras.Input(shape = rpnetwork_size)
        
        #update backbone
        backbone_output = self.backbone(backbone_input)
        self.backbone = tf.keras.Model(backbone_input, backbone_output)
        
        #update rpnetwork
        rpnetwork_output = self.rpnetwork(rpnetwork_input)
        self.rpnetwork = tf.keras.Model(rpnetwork_input, rpnetwork_output)
        
        
    @tf.function
    def threshold(self, boxes, objectness, tau):
        """
        Thresholds rpn predictions using objectness score. Helpful for processing
        raw inference results.
        
        Parameters
        ----------
        boxes: tensor
            N x 4 tensor where each row contains the x,y location of the upper left
            corner of a regressed box and its width and height in that order.
        objectness:
            N x 2 tensor containing corresponding softmax objectness scores in rows.
            Second column contains score for being an object.
        tau: float
            Scalar threshold applied to objectness scores to define which rpn 
            predictions are objects. Range is [0,1].
        
        Returns
        -------
        filtered_boxes: tensor
            M x 4 tensor containing objectness score filtered boxes (M <= N).
        filtered_objectness tensor
            M x 2 tensor containing objectness score filtered objectness scores (M <= N).       
        """
        
        #get binary mask of positive objects
        mask = tf.greater(objectness[:,1], tau)
        
        #filter using mask
        filtered_boxes = tf.boolean_mask(boxes, mask, axis=0)
        filtered_objectness = tf.boolean_mask(objectness, mask, axis=0)       
        
        return filtered_boxes, filtered_objectness, mask
        
        
    @tf.function
    def nms(self, boxes, objectness, nms_iou):
        """Performs nms on regressed boxes returning filtered boxes and corresponding
        objectness scores. Helpful for processing raw inference results.
        
        Parameters
        ----------
        boxes: tensor
            N x 4 tensor where each row contains the x,y location of the upper left
            corner of a regressed box and its width and height in that order.
        objectness: tensor
            N x 2 tensor containing corresponding softmax objectness scores in rows.
            Second column contains score for being an object.
        nms_iou: float
            Scalar threshold used by nms to define overlapping objects. Range is
            (0,1].
            
        Returns
        -------
        nms_boxes: tensor
            M x 4 tensor containing nms filtered boxes (M <= N).
        nms_objectness tensor
            M x 2 tensor containing nms filtered objectness scores (M <= N).
        """    
        
        #get indices of boxes selected by NMS
        selected = tf.image.non_max_suppression(tf_box_transform(boxes),
                                                objectness[:,1], tf.shape(objectness)[0],
                                                iou_threshold=nms_iou)
        
        #filter intputs to discard unselected boxes
        nms_boxes = tf.gather(boxes, selected, axis=0)
        nms_objectness = tf.gather(objectness, selected, axis=0)
    
        return nms_boxes, nms_objectness, selected
    
    
    @tf.function
    def align(self, boxes, features, field, pool, tiles):
        """Performs roialign on filtered inference results. If results are not filtered
        using objectness thresholding or non-max suppression prior to this step an OOM 
        error may occur. Helpful for refinement of post-processed inference results from 
        raw.
        
        Parameters
        ----------
        features: tensor
            The three dimensional feature map tensor produced by the backbone network.
        boxes: tensor
            N x 4 tensor where each row contains the x,y location of the upper left
            corner of an rpn regressed box and its width and height in that order.
        field: float32
            Field size of the backbone network in pixels.
        pool: int32
            pool^2 is the number of locations to interpolate features at within 
            each tile.
        tiles: int32
            tile^2 is the number of tiles that each regressed bounding box is divided
            into.
            
        Returns
        -------
        align_boxes: tensor (float32)
            N x 4 tensor containing aligned boxes.
        nms_objectness: tensor (float32)
            N x 2 tensor containing corresponding objectness scores.
        """
 
        #interpolate features in pool*tiles x pool*tiles grid for each box
        interpolated = roialign(features, boxes, field, pool, tiles)
        
        #generate roialign predictions and transform to box representation
        align_reg = self.fastrcnn(interpolated)
        align_boxes = unparameterize(align_reg, boxes)
        
        return align_boxes
    
    
    @tf.function
    def raw(self, rgb):
        """
        raw() produces unfiltered objectness scores and regressions from the rpn
        network, and backbone features. Additional steps are required for thresholding
        based on scores, nms, and performing roialign. This is useful for users who
        would like to provide their own post-processing of rpn results.
        
        Parameters
        ----------
        rgb: tensor (uint8)
            An M x N x 3 image to perform inference on.
       
        Returns
        -------
        rpn_obj: tensor (float32)
            D x 2 tensor containing objectness scores from the region proposal network.
        rpn_boxes: tensor (float32)
            D x 4 tensor containing regressions from the region proposal network.
        features: tensor (float32)
            An M x N x K tensor of feature maps generated by the backbone. Can be used 
            to subsequently apply RoiAlign or other operations.
        """

        #expand dimensions
        rgb = tf.expand_dims(rgb, axis=0)
        
        #predict and capture intermediate features
        features = self.backbone(rgb, training=False)
        output = self.rpnetwork(features, training=False)

        #generate anchors for input size image
        anchors = create_anchors(self.anchor_px, self.field, 
                                 tf.shape(rgb)[2], tf.shape(rgb)[1])

        #transform outputs to 2D arrays with anchors in rows
        rpn_obj = tf.nn.softmax(map_outputs(output[0], anchors,
                                self.anchor_px, self.field))
        rpn_reg = map_outputs(output[1], anchors, self.anchor_px, self.field)        
        rpn_boxes = unparameterize(rpn_reg, anchors)
        
        #clip regressed boxes to border
        rpn_boxes = clip_boxes(rpn_boxes, tf.shape(rgb)[2], tf.shape(rgb)[1])
        
        return rpn_obj, rpn_boxes, features
    
    
    @tf.function
    def call(self, rgb, tau=None, nms_iou=None, margin=0):
        """
        call() produces thresholded and roialign refined predictions from a trained
        network. This is the most useful for users who don't want to apply their own
        post-processing to rpn results.
        
        Parameters
        ----------
        rgb: tensor (uint8)
            An M x N x 3 image to perform inference on.
        tau: float32
            Threshold in range [0,1] used to select region proposals based on 
            region proposal network objectness scores. Defaults to model attribute 
            (default value 0.5).
        nms_iou: float32
            Intersection over union threshold used to remove redundant proposals
            during non-max suppression. Range is (0, 1]. Defaults to model
            attribute (default value 0.3).
        margin: int32
            The margin value is used to clear predictions from the edge of the
            input image. All objects intersecting partially or wholly with this 
            margin will be removed. Default value 0 pixels performs no clearing.
            
        Returns
        -------
        align_boxes: tensor (float32)
            A D x 4 tensor of objectness thresholded, non-max suppressed, and 
            RoiAligned regoin proposals.
        """
        
        #generate raw rpn outputs
        rpn_obj, rpn_boxes, features = self.raw(rgb)
        
        #select rpn proposals
        if tau is None:
            tau = self.tau
        rpn_boxes_positive, rpn_obj_positive, positive = self.threshold(rpn_boxes, rpn_obj, 
                                                                        tau)
        
        #perform non-max suppression on rpn positive predictions
        if nms_iou is None:
            nms_iou = self.nms_iou
        rpn_boxes_nms, rpn_obj_nms, selected = self.nms(rpn_boxes_positive,
                                                        rpn_obj_positive, nms_iou)
        
        #generate roialign predictions for rpn positive predictions
        align_boxes = self.align(rpn_boxes_nms, features, self.field, 
                                 self.pool, self.tiles)
        
        #filter edge boxes
        filtered, mask = filter_edge_boxes(align_boxes, tf.shape(rgb)[1],
                                           tf.shape(rgb)[0], margin)
        
        return filtered

        
    @tf.function(experimental_relax_shapes=True)
    def test_step(self, data):
        """
        Test step performs inference on a single image, calculating both single-image
        and aggregate metrics. Single-image metrics are printed to the stdout.
        """
        
        #unpack image, boxes, and optional image name
        if len(data) ==3:
            rgb, boxes, name = data
        else:
            rgb, boxes = data
            name = ''

        #convert boxes from RaggedTensor
        boxes = boxes.to_tensor()
        
        #generate rpn predictions
        rpn_obj, rpn_boxes, features = self.raw(rgb)
     
        #evaluate rpn - generate anchors for input size image
        anchors = create_anchors(self.anchor_px, self.field, 
                                 tf.shape(rgb)[1], tf.shape(rgb)[0])
        
        #evaluate rpn - filter into positive and negative anchors
        positive_anchors, negative_anchors = filter_anchors(boxes, anchors)

        #evaluate rpn - generate objectness scores
        output = self.rpnetwork(features, training=False)[0]
 
        #evaluate rpn - calculate objectness predictions and labels
        positive_obj = tf.nn.softmax(map_outputs(output, positive_anchors,
                                                 self.anchor_px, self.field))
        negative_obj = tf.nn.softmax(map_outputs(output, negative_anchors,
                                                 self.anchor_px, self.field))
        obj_labels = tf.concat([tf.ones(tf.shape(positive_obj)[0], tf.uint8),
                                tf.zeros(tf.shape(negative_obj)[0], tf.uint8)],
                                axis=0)
        
        #select positive rpn proposals for nms and roialign
        rpn_boxes_pred, rpn_obj_pred, _ = self.threshold(rpn_boxes, rpn_obj, self.tau)
        
        #perform non-max suppression on boxes
        rpn_boxes_nms, rpn_obj_nms, selected = self.nms(rpn_boxes_pred,
                                                        rpn_obj_pred, self.nms_iou)
        
        #generate roialign predictions for rpn positive predictions
        align_boxes_nms = self.align(rpn_boxes_nms, features, self.field, self.pool, self.tiles)
        
        #clear margin of ground truth boxes
        boxes, _ = filter_edge_boxes(boxes, tf.shape(rgb)[1], tf.shape(rgb)[0], self.margin)     
        
        #filter edge boxes
        filtered, mask = filter_edge_boxes(align_boxes_nms, tf.shape(rgb)[1],
                                           tf.shape(rgb)[0], self.margin)
        filtered_objectness = tf.boolean_mask(rpn_obj_nms, mask, axis=0)
        
        #calculate ious
        align_ious = iou(filtered, boxes)
        
        #greedy iou mapping for precision-recall auc
        tp, fp, fn, tp_list, fp_list, fn_list = greedy_iou_mapping(align_ious, self.tpr_iou)
        
        #update console
        tf.print(name)
        tf.print('greedy tp: ', tp)
        tf.print('greedy fp: ', fp)
        tf.print('greedy fn: ', fn)
        
        #best iou with ground truth for each prediction - reduce along rows
        align_ious = tf.reduce_max(align_ious, axis=1)
        
        #update objectness binary classification metrics
        obj_metrics = self._update_objectness_metrics(tf.concat([positive_obj,
                                                                 negative_obj],
                                                                axis=0),
                                                      obj_labels)
        
        #update regression metrics
        reg_metrics = self._update_regression_metrics(boxes,
                                                      tf.concat([filtered_objectness,
                                                                 filtered],
                                                                axis=1))
        
        #combine metrics
        metrics = {**obj_metrics, **reg_metrics}
        
        #dummy losses
        losses = {'loss_rpn_obj': 0., 'loss_rpn_reg': 0., 'loss_align_reg': 0.}        
        
        return {**losses, **metrics}
    
    
    @tf.function
    def train_step(self, data):
        """
        Train step performs gradient updates on single image batches. Aggregate metrics are
        calculated.
        """
    
        #unpack image, boxes, and optional image name
        if len(data) ==3:
            rgb, boxes, name = data
        else:
            rgb, boxes = data

        #convert boxes from RaggedTensor
        boxes = boxes.to_tensor()

        #expand dimensions
        norm = tf.expand_dims(tf.cast(rgb, tf.float32), axis=0)

        #label anchors as negative, positive
        positive_anchors, negative_anchors = filter_anchors(boxes, self.anchors)

        #training step
        with tf.GradientTape(persistent=True) as tape:
            
            #predict and capture intermediate features
            features = self.backbone(norm, training=True)
            output = self.rpnetwork(features, training=True)
            
            #calculate rpn objectness loss for all positive and negative anchors
            positive_obj = tf.nn.softmax(map_outputs(output[0], positive_anchors,
                                                         self.anchor_px, self.field))
            negative_obj = tf.nn.softmax(map_outputs(output[0], negative_anchors,
                                                         self.anchor_px, self.field))
            
            #sample anchors and perform hard-negative mining
            positive_anchors, negative_anchors = sample_anchors(positive_anchors, 
                                                                negative_anchors,
                                                                negative_obj[:,1],
                                                                self.max_anchors, 
                                                                self.np_ratio,
                                                                self.hard_fraction)

            #calculate objectness loss for sampled positive and negative anchors
            positive_obj = tf.nn.softmax(map_outputs(output[0], positive_anchors,
                                                         self.anchor_px, self.field))
            negative_obj = tf.nn.softmax(map_outputs(output[0], negative_anchors,
                                                         self.anchor_px, self.field))
            obj_labels = tf.concat([tf.ones(tf.shape(positive_obj)[0], tf.uint8),
                                        tf.zeros(tf.shape(negative_obj)[0], tf.uint8)],
                                       axis=0)
            obj_loss = self.loss[0](tf.concat([positive_obj, negative_obj], axis=0),
                                    tf.one_hot(obj_labels, 2))

            #ccalculate rpn regression loss
            regression = map_outputs(output[1], positive_anchors, self.anchor_px,
                                     self.field)
            reg_label = parameterize(positive_anchors, boxes)
            reg_loss = self.loss[1](reg_label, regression)          

            #weighted sum of objectness and regression losses
            rpn_loss = obj_loss / self.max_anchors + \
                reg_loss * self.lmbda / tf.cast(tf.shape(self.anchors)[0], tf.float32)
            
            #roi align and fast rcnn operations to be performed if positive anchors exist
            def nonempty():
            
                #fast r-cnn regression of roialign results for positive anchors
                rpn_boxes = unparameterize(regression, positive_anchors)
                rpn_boxes_positive, _ = filter_anchors(boxes, rpn_boxes)
                interpolated = roialign(features, rpn_boxes_positive, self.field, 
                                        pool=self.pool, tiles=self.tiles)
                align = self.fastrcnn(interpolated)
            
                #transform aligned regressions to box representation and calculate labels
                align_boxes = unparameterize(align, rpn_boxes_positive)
                align_label = parameterize(rpn_boxes_positive, boxes)
            
                return self.loss[1](align_label, align)
        
            def empty():
                return 0.0
        
            #conditional when positive anchors exist
            align_loss = tf.cond(tf.size(positive_anchors) > 0, nonempty, empty)
                
        #calculate backbone gradients and optimize
        gradients = tape.gradient(rpn_loss, self.backbone.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.backbone.trainable_weights))

        #calculate rpn gradients and optimize
        gradients = tape.gradient(rpn_loss, self.rpnetwork.trainable_weights)  
        self.optimizer.apply_gradients(zip(gradients, self.rpnetwork.trainable_weights))
            
        #calculate roialign gradients and optimize
        gradients = tape.gradient(align_loss, self.fastrcnn.trainable_weights)  
        self.optimizer.apply_gradients(zip(gradients, self.fastrcnn.trainable_weights))
        
        #update metrics
        metrics = self._update_objectness_metrics(tf.concat([positive_obj,
                                                             negative_obj], 
                                                            axis=0),
                                                  obj_labels)

        #build output loss dict
        losses = {'loss_rpn_obj': obj_loss, 'loss_rpn_reg': reg_loss,
                  'loss_align_reg': align_loss}

        return {**losses, **metrics}



import numpy as np


def field_size(backbone, length=65):
    """Estimates receptive field size of the backbone network.
    
    The receptive field size is estimated empirically by passing arrays of increasing
    sizes through the backbone and comparing the ratio of sizes of the input and output.
    The max of this sequence of ratios corresponds to the case where the field size
    evenly divides the input with no remainder.
        
    Parameters
    ----------
    backbone: keras Model
        The backbone network as a keras Model.
    length: int32
        Maximum size of input to use in estimating field size.
        
    Returns
    -------
    field: float32
        Field size of the backbone network in pixels.
    """    
  
    ratio = []
    for dim in np.arange(1, length, 1):
        im = np.zeros(shape=(dim, dim, 3), dtype=np.uint8)
        prediction = backbone.predict(np.array([im,])).shape
        ratio.append(dim/prediction[2])
    field = np.max(ratio)

    return field



"""
This package contains functions for generating, filtering, and sampling anchors
"""

# make functions available at the package level using shadow imports
from .create import create_anchors
from .filter import filter_anchors
from .sampling import sample_anchors

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'create_anchors',
    'filter_anchors',
    'sample_anchors'
)



from histomics_detect.metrics.iou import iou
import tensorflow as tf


@tf.function
def filter_anchors(boxes, anchors, alpha=0.7, beta=0.3, gamma=0.3):
    """Compares anchors to ground truth to determine positive and negative
    anchors.
    
    This function uses multiple criteria for intersection-over-union (IoU) 
    with ground truth to label anchors as either positive or negative. 
    A positive anchor is any anchor that has over 'alpha' IoU with any ground 
    truth object, or that is the max IoU anchor for a ground truth object 
    with IoU at least 'beta'. A negative anchor is any anchor where the max 
    IoU with ground truth objects is less than 'gamma'.
        
    Parameters
    ----------
    boxes: tensor (float32)
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    anchors: tensor (float32)
        M x 4 tensor anchor positions organized in a dense grid over the image
        space. Each row contains the x,y upper left corner of the anchor in pixel 
        units relative in the image coordinate frame, and the anchor width and 
        height.
    alpha: float32
        Threshold for calling an anchor positive based on maximum IoU with any
        ground truth object.
    beta: float32
        Threshold for calling an anchor positive based on being the highest IoU
        anchor for a single ground truth object.
    gamma: float32
        Threshold for calling anchors negative.
        
    Returns
    -------
    positive: tensor (float32)
        M x 4 tensor of positive anchor positions satisfying IoU criteria
        with some ground truth object. Each row contains the x,y upper left corner 
        of the anchor in pixel units relative in the image coordinate frame, and
        the anchor width and height.
    negative: tensor (float32)
        N x 4 tensor of negative anchor positions not satisfying IoU criteria with
        any ground truth object. Each row contains the x,y upper left corner of the
        anchor in pixel units relative in the image coordinate frame, and the
        anchor width and height.
    """

    #calculate IOU between ground truth and anchors
    ious = iou(boxes, anchors)

    #anchors where IOU > alpha
    matches = tf.cast(tf.where(tf.greater(ious, alpha)), dtype=tf.int32)
    alpha_id = matches[:,1]
    alpha_box = matches[:,0]

    #anchors where max match for a box and IOU > beta
    beta_max = tf.reduce_max(ious, axis=1)
    beta_id = tf.argmax(ious, axis=1, output_type=tf.int32)
    beta_box = tf.range(0,tf.shape(boxes)[0], dtype=tf.int32)

    #filter out duplicates
    keep = tf.where(tf.logical_and(tf.greater(beta_max, beta),
                                   tf.less(beta_max, alpha)))[:,0]
    beta_id = tf.gather(beta_id, keep)
    beta_box = tf.gather(beta_box, keep)

    #combine positive anchors
    alpha_anchors = tf.concat([tf.gather(anchors, alpha_id),
                               tf.expand_dims(tf.cast(alpha_box, tf.float32), axis=1)],
                              axis=1)
    beta_anchors = tf.concat([tf.gather(anchors, beta_id),
                              tf.expand_dims(tf.cast(beta_box, tf.float32), axis=1)],
                             axis=1)
    positive = tf.concat([alpha_anchors, beta_anchors], axis=0)

    #anchors with IOU < gamma
    negative = tf.gather(anchors, tf.where(tf.less(tf.reduce_max(ious, axis=0), gamma)))
    negative = tf.squeeze(negative)

    return positive, negative



"""
This package contains functions for region of interest alignment
"""

# make functions available at the package level using shadow imports
from .roialign import roialign

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'roialign'
)



"""
This package contains functions for visualization of detection results
"""

# make functions available at the package level using shadow imports
from .visualization import plot_evaluation
from .visualization import plot_inference
from .lnms_visualization import plot_inference as plot_inference_lnms, run_plot, plot_multiple_outputs

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'plot_evaluation',
    'plot_inference',
    'plot_inference_lnms',
    'run_plot',
    'plot_multiple_outputs'
)



"""
This package contains performance metrics implemented in tensorflow
"""

# make functions available at the package level using shadow imports
from .average_precision import AveragePrecision
from .iou import iou
from .iou import greedy_iou_mapping
from .lnms import tf_linear_sum_assignment
from .objectness import FalseNegativeRate, FalsePositiveRate

__all__ = (
    'AveragePrecision',
    'iou',
    'greedy_iou_mapping',
    'tf_linear_sum_assignment',
    'FalseNegativeRate',
    'FalsePositiveRate'
)



from abc import ABC

import tensorflow as tf


class CompressionNetwork(tf.keras.Model, ABC):

    def __init__(self, feature_size: int, anchor_size: int, backbone: tf.keras.Model, *args, **kwargs):
        """
        Network for training the compression layers of for the lnms network

        Parameters
        ----------
        feature_size: int
            compression size of the feature
        anchor_size: int
        backbone: tf.keras.Model
            feature extraction model
        args
        kwargs
        """
        super(CompressionNetwork, self).__init__(*args, **kwargs)
        self.feature_size = feature_size
        self.anchor_size = anchor_size
        self.activation = 'sigmoid'
        self.compression_layers = self._initialize_compression_layers()
        self.decompression_layers = self._initialize_decompression_layers()
        self.backbone = backbone

    def _initialize_decompression_layers(self) -> tf.keras.Model:
        """
        builds the initial feature decompression keras model

        Returns
        -------
        compression_model: tf.keras.Model
            compression network that can compress image features
        """
        input_layer = tf.keras.Input(shape=[None, None, self.feature_size], name="compressed_input")
        layer_1 = tf.keras.layers.Conv2DTranspose(self.feature_size, 1, activation=self.activation, padding='same',
                                                  name="decompression_layer_1")(input_layer)
        layer_2 = tf.keras.layers.Conv2DTranspose(self.anchor_size, 1, activation=self.activation, padding='same',
                                                  name="decompression_layer_2")(layer_1)
        return tf.keras.Model(inputs=input_layer, outputs=layer_2, name="decompression_network")

    def _initialize_compression_layers(self) -> tf.keras.Model:
        """
        builds the initial feature compression keras model

        Returns
        -------
        compression_model: tf.keras.Model
            compression network that can compress image features
        """
        input_layer = tf.keras.Input(shape=[None, None, self.anchor_size], name="uncompressed_input")
        layer_1 = tf.keras.layers.Conv2D(self.feature_size, 3, activation=self.activation, padding='same',
                                         name="compression_layer_1")(input_layer)
        layer_2 = tf.keras.layers.Conv2D(self.feature_size, 3, activation=self.activation, padding='same',
                                         name="compression_layer_2")(layer_1)
        return tf.keras.Model(inputs=input_layer, outputs=layer_2, name="compression_network")

    def call(self, inputs, training=None, mask=None):
        return self.decompression_layers(self.compression_layers(inputs))

    @tf.function
    def train_step(self, data):
        (img, boxes, name) = data
        # normalize image
        norm = tf.keras.applications.resnet.preprocess_input(tf.cast(img, tf.float32))
        # expand dimensions
        norm = tf.expand_dims(norm, axis=0)

        features = self.backbone(norm)

        with tf.GradientTape(persistent=True) as tape:
            network_output = self(features)
            loss = self.loss(network_output, features)

        gradients = tape.gradient(loss, self.compression_layers.trainable_weights +
                                  self.decompression_layers.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.compression_layers.trainable_weights +
                                           self.decompression_layers.trainable_weights))

        # save loss and metrics
        losses = {'loss': loss}
        metrics = {}

        return {**losses, **metrics}



from abc import ABC
from typing import Tuple
import tensorflow as tf

from histomics_detect.networks.field_size import field_size
from histomics_detect.anchors.create import create_anchors
from histomics_detect.models.block_model import BlockModel
from histomics_detect.roialign.roialign import roialign
from histomics_detect.boxes.transforms import unparameterize, parameterize, clip_boxes
from histomics_detect.models.lnms_loss import normal_loss, clustering_loss, paper_loss, xor_loss, normal_clustering_loss
from histomics_detect.boxes.match import cluster_assignment
from histomics_detect.boxes.cross_boxes import cross_from_boxes


class LearningNMS(tf.keras.Model, ABC):
    def __init__(self, configs: dict, rpnetwork: tf.keras.Model, backbone: tf.keras.Model,
                 compression_network: tf.keras.Model, shape, *args, **kwargs):
        """
        Learning-NMS model for training a Block-Model

        Parameters
        ----------
        configs: dictionary
            configuration dict
        rpnetwork: tf.keras.Model
            rpnetwork for extracting the boxes from the image
        backbone: tf.keras.Model
            network to extract image features
        shape:
        args
        kwargs
        """
        super(LearningNMS, self).__init__(*args, **kwargs)
        self.configs = configs

        for key, value in configs.items():
            setattr(self, key, value)

        for key, value in kwargs.items():
            setattr(self, key, value)

        # add models to self
        self.rpnetwork = rpnetwork
        self.backbone = backbone
        self.lmbda = configs['rpn_lmbda']

        self.field = field_size(self.backbone)
        self.anchors = create_anchors(self.anchor_px, self.field, shape[0], shape[1])

        self.compression_net = compression_network

        if not self.data_only:
            self.net = BlockModel([self._initialize_block(i) for i in range(self.num_blocks)],
                                  self._initialize_final_output(), threshold=self.threshold,
                                  train_tile=self.train_tile, use_image_features=self.use_image_features,
                                  use_distance=self.use_distance, original_lnms=self.original_lnms)

            if self.use_reg:
                self.init_regression = self._initialize_init_regression()

            # define metrics
            loss_col = tf.keras.metrics.Mean(name='loss')
            loss_pos = tf.keras.metrics.Mean(name='pos_loss')
            loss_neg = tf.keras.metrics.Mean(name='neg_loss')
            self.standard = [loss_col, loss_pos, loss_neg]

    def _initialize_init_regression(self) -> tf.keras.Model:
        feature_size_multiplier = 2 if self.combine_box_and_cross else 1
        shape = (feature_size_multiplier * self.feature_size + 5)

        init_input = tf.keras.Input(shape=shape, name="init_regression_input")

        layer1 = tf.keras.layers.Dense(int(shape / 8), activation=self.activation, use_bias=True,
                                       name=f'init_regression_layer_1')(init_input)
        layer2 = tf.keras.layers.Dense(int(shape / 8), activation=self.activation, use_bias=True,
                                       name=f'init_regression_layer_2')(layer1)
        layer3 = tf.keras.layers.Dense(4, activation='linear', use_bias=True,
                                       name=f'init_regression_layer_3')(layer2)

        return tf.keras.Model(inputs=init_input, outputs=layer3, name="init_regression_network")

    def _initialize_final_output(self) -> tf.keras.Model:
        """
        builds the final layers of the lnms network

        'num_hidden_layers' Dense layers with 'final_hidden_layer_features' per layer
        output dimension is 1

        Returns
        -------
        final_output: tf.keras.Model
            network for compressing neighborhood to single prediction vector
        """
        feature_size_multiplier = 2 if self.combine_box_and_cross and self.use_image_features else 1

        final_input = tf.keras.Input(shape=self.feature_size * feature_size_multiplier + 1,
                                     name="final_after_block_layers_input")
        x = final_input
        for i in range(self.num_hidden_layers):
            x = tf.keras.layers.Dense(self.final_hidden_layer_features, activation=self.activation, use_bias=True,
                                      name=f'final_after_block_layer_{i}')(x)
        output_size = self.add_regression_param * 2 + 1 + int(self.objectness_format)
        x = tf.keras.layers.Dense(output_size, activation=self.final_activation, name="output_score_layer",
                                  use_bias=True)(x)

        return tf.keras.Model(inputs=final_input, outputs=x)

    def _initialize_block(self, block_id: int):
        """
        builds two keras models for each block

        block:
            - block: 'num_layers_block' hidden dense layers with 'block_hidden_layer_features' features for
                each layer
                processes a neighborhood at a time

                input size: [N x 2*D]
                hidden size: [N x 'block_hidden_layer_features']
                output size: hidden size

            - output: single dense layer that maps the pooled neighborhood to the size of a single feature

                input size: [N x 'block_hidden_layer_features']
                output size: [1 x D]

        Parameters
        ----------
        block_id: int
            number of the current block

        Returns
        -------
        block, output: tf.keras.Model, tf.keras.Model
            block keras model, output keras model
        """
        feature_size_multiplier = 2 if self.combine_box_and_cross and self.use_image_features else 1

        shape = 6 + (feature_size_multiplier * 2 * self.feature_size + 2)
        block_input = tf.keras.Input(shape=shape, name=f'block_{block_id}_input')
        x = tf.keras.layers.BatchNormalization(axis=1, name=f'block_{block_id}_batch_norm_0_layer')(block_input)
        for i in range(self.num_layers_block):
            x = tf.keras.layers.Dense(self.block_hidden_layer_features, activation=self.activation,
                                      name=f'block_{block_id}layer{i}', use_bias=True)(x)
        normed_x = tf.keras.layers.BatchNormalization(axis=1, name=f'block_{block_id}_batch_norm_1_layer')(x)
        block = tf.keras.Model(inputs=block_input, outputs=normed_x, name=f'block_{block_id}_layers')

        final_input = tf.keras.Input(shape=self.block_hidden_layer_features, name=f'block_{block_id}_after_pool_input')
        output = tf.keras.Model(inputs=final_input,
                                outputs=tf.keras.layers.Dense(self.feature_size * feature_size_multiplier + 1,
                                                              activation='linear',
                                                              name=f'block_{block_id}_after_pool_layer', use_bias=True)(
                                    final_input),
                                name=f'block_{block_id}_output')

        return block, output

    def _interpolate_features(self, features: tf.Tensor, rpn_boxes: tf.Tensor) -> tf.Tensor:
        """
        interpolate features for the given boxes

        Parameters
        ----------
        features: tensor (float32)
            features extracted from the image
        rpn_boxes: tensor (float32)
            shape: N x 4
            region proposal boxes

        Returns
        -------
        interpolated: tensor (float32)
            shape: N x s
            interpolated feature for each box
        """
        reduction_func = tf.reduce_mean if self.reduce_mean else tf.reduce_max

        # calculate interpolated features
        if self.cross_boxes:
            cross_boxes = cross_from_boxes(rpn_boxes, self.cross_scale, image_width=self.width,
                                           image_height=self.height)
            interpolated = roialign(features, tf.reshape(cross_boxes, (-1, 4)), self.field,
                                    pool=self.roialign_pool, tiles=self.roialign_tiles)
            interpolated = reduction_func(interpolated, axis=1)
            interpolated = reduction_func(interpolated, axis=1)
            interpolated = tf.reshape(interpolated, (tf.shape(rpn_boxes)[0], 2, -1))
            interpolated = reduction_func(interpolated, axis=1)

            if self.combine_box_and_cross:
                interpolated_box = roialign(features, rpn_boxes, self.field,
                                            pool=self.roialign_pool, tiles=self.roialign_tiles)
                interpolated_box = reduction_func(interpolated_box, axis=1)
                interpolated_box = reduction_func(interpolated_box, axis=1)
                interpolated = tf.concat([interpolated, interpolated_box], axis=1)
            # TODO maybe concatenate instead of mean for cross
        else:
            if self.expand_boxes:
                rpn_boxes = tf.concat(
                    [rpn_boxes[:, :2] - self.box_expand_value, rpn_boxes[:, 2:] + self.box_expand_value * 2])
                rpn_boxes = clip_boxes(rpn_boxes, self.width, self.height)
            interpolated = roialign(features, rpn_boxes, self.field,
                                    pool=self.roialign_pool, tiles=self.roialign_tiles)
            interpolated = reduction_func(interpolated, axis=1)
            interpolated = reduction_func(interpolated, axis=1)

        interpolated = tf.reshape(interpolated, [tf.shape(interpolated)[0], -1])

        return interpolated

    def call(self, data, training=None, mask=None):
        features, boxes, rpn_boxes, scores = data

        compressed_features = self.compression_net(features, training=False)

        interpolated = self._interpolate_features(compressed_features, rpn_boxes)
        interpolated = tf.concat([scores, interpolated], axis=1)

        # run network
        nms_output = self.net((interpolated, rpn_boxes), training=True)
        return features, rpn_boxes, scores, nms_output

    def train_step(self, data):
        """
        training function

        Parameters
        ----------
        data: Tuple[features, boxes, rpn_boxes, scores]
            - features: tensor (float32)
                shape: W x H x M
                features extracted from the image
            - boxes: tensor (float32)
                shape: N x 4
                ground truth boxes
            - rpn_boxes: tensor (float32)
                shape: G x 4
                region proposal boxes
            - scores:
                shape: G x 1
                objectness scores

        Returns
        -------
        losses, metrics
        """
        features, boxes, rpn_boxes, scores = data

        if not self.compressed_gradient:
            compressed_features = self.compression_net(features, training=False)

        if not self.interpolated_gradient:
            # calculate interpolated features
            interpolated = self._interpolate_features(compressed_features, rpn_boxes)
            interpolated_score = tf.concat([scores, interpolated], axis=1)

        if self.use_reg:
            # TODO does not work fix or remove
            rpn_reg_label = parameterize(rpn_boxes, boxes)

        # training step
        with tf.GradientTape(persistent=True) as tape:

            if self.compressed_gradient:
                compressed_features = self.compression_net(features, training=True)

            if self.interpolated_gradient:
                # calculate interpolated features
                interpolated = self._interpolate_features(compressed_features, rpn_boxes)
                interpolated_score = tf.concat([scores, interpolated], axis=1)

            if self.use_reg:
                init_regression_input = tf.concat([interpolated_score, rpn_boxes], axis=1)
                init_regression_output = self.init_regression(init_regression_input)

                rpn_boxes = unparameterize(init_regression_output, rpn_boxes)
                rpn_reg_loss = self.loss[1](rpn_reg_label, init_regression_output)

            if self.manipulate_rpn:
                rpn_boxes, scores = self.maipulate_rpn_func(self, rpn_boxes, boxes, interpolated, scores)
                interpolated_score = tf.concat([scores, interpolated], axis=1)

            # run network
            nms_output = self.net((interpolated_score, rpn_boxes), training=True)

            loss, labels = self._calculate_loss(nms_output, boxes, rpn_boxes)

        gradients = tape.gradient(loss, self.net.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.net.trainable_weights))

        if self.use_reg:
            gradients = tape.gradient(rpn_reg_loss, self.init_regression.trainable_weights)
            self.optimizer.apply_gradients(zip(gradients, self.init_regression.trainable_weights))

        if self.use_image_features and self.compressed_gradient:
            gradients = tape.gradient(loss, self.compression_net.trainable_weights)
            self.optimizer.apply_gradients(zip(gradients, self.compression_net.trainable_weights))

        self.standard[0].update_state(loss + 1e-8)

        if self.calculate_train_metrics:
            self._cal_update_performance_stats(boxes, rpn_boxes, nms_output)

        # save loss and metrics
        losses = {'lnms_loss': loss}
        metrics = {m.name: m.result() for m in self.standard}

        return {**losses, **metrics}

    def _calculate_loss(self, nms_output, boxes, rpn_boxes) -> Tuple[float, tf.Tensor]:
        """
        call the loss function of the loss specified in 'loss_type'

        S: size of neighborhood
        N: number of predictions
        D: size of a single prediction
        G: number of ground truth boxes

        Parameters
        ----------
        nms_output: tensor (float32)
            objectiveness scores corresponding to the predicted boxes after lnms processing
            shape: N x 1
        boxes tensor (float32)
            ground truth boxes
            shape: G x 4
        rpn_boxes: tensor (float32)
            shape: N x 4
            each prediction in box form

        Returns
        -------
        loss
        labels
        """
        # calculate loss
        if self.loss_type == 'dummy':
            scores = tf.expand_dims(nms_output[:, 0], axis=1)
            loss = tf.reduce_sum(self.loss_object(scores, tf.ones(tf.shape(nms_output))))
            labels = []
        elif self.loss_type == 'xor':
            scores = tf.expand_dims(nms_output[:, 0], axis=1)
            clusters = cluster_assignment(boxes, rpn_boxes)
            loss, labels = xor_loss(scores, clusters)
            # loss, labels = self._cal_xor_loss(nms_output, cluster_assignment)
        elif self.loss_type == 'clustering':
            clusters = cluster_assignment(boxes, rpn_boxes)
            loss, labels = clustering_loss(nms_output, clusters, self.loss_object, self.positive_weight,
                                           self.standard, boxes, rpn_boxes, self.weighted_loss, self.neg_pos_loss,
                                           self.add_regression_param)
        elif self.loss_type == 'paper':
            scores = tf.expand_dims(nms_output[:, 0], axis=1)
            loss, labels = paper_loss(boxes, rpn_boxes, scores, self.loss_object, self.positive_weight,
                                      self.standard, self.weighted_loss, self.neg_pos_loss, self.iou_threshold)
        elif self.loss_type == 'clustering_normal':
            clusters = cluster_assignment(boxes, rpn_boxes)
            loss, labels = normal_clustering_loss(nms_output, boxes, rpn_boxes, clusters, self.loss_object,
                                                  self.positive_weight, self.standard, self.weighted_loss,
                                                  self.neg_pos_loss, self.use_pos_neg_loss, self.norm_loss_weight,
                                                  self.add_regression_param, self.iou_threshold)
        elif self.loss_type == 'custom':
            loss, labels = self.custom_loss(self, nms_output, boxes, rpn_boxes)
        else:
            scores = tf.expand_dims(nms_output[:, 0], axis=1)
            loss, labels = normal_loss(self.loss_object, boxes, rpn_boxes, scores, self.positive_weight,
                                       self.standard, neg_pos_loss=True, min_iou=self.iou_threshold)

        return loss, labels



import tensorflow as tf


from histomics_detect.augmentation.augmentation import _box_crop


def cross_from_boxes(boxes: tf.Tensor, scale: float, width: int = 10, height: int = 10, image_width: int = 224,
                     image_height: int = 224, grow: bool = False) -> tf.Tensor:
    """
    generates cross shape for each box (similar to CornerNet)

    takes the center (c_x, c_y) of the boxes (x, y, w, h) and creates two boxes from that:
    - width box (vertical line of the cross):
        spans from (c_x-'width'/2, c_y-'factor'*h) to (c_x+'width'/2, c_y+'factor'*h)
    - height box (horizontal line of the cross):
        spans from (c_x-'factor'*w, c_y-'height'/2) to (c_x+'factor'*w, c_y+'height'/2)

    the cross is capped such that it is entirely within the image boundaries

    N: number of boxes

    Parameters
    ----------
    boxes: tensor (int32)
        input boxes to be turned into crosses
        shape (N, 4)
    scale: float
        see above

        scale * box_width, scale * box_height is the width and height of the cross
    width: int
        width of the vertical line
    height: int
        height of the horizontal line
    image_width: int
        width of the image
    image_height: int
        height of the image
    grow: bool
        make cross always span the whole image

    Returns
    -------
    cross_boxes: tensor (float32)
        vertical and horizontal boxes of each cross for each box
        shape: (N, 2, 4)
    """

    image_width = tf.cast(image_width, tf.float32)
    image_height = tf.cast(image_height, tf.float32)

    centers = boxes[:, :2] + boxes[:, 2:]/2

    if grow:
        horizontal_width = image_width * tf.ones(tf.shape(boxes[:, 2]))
        vertical_width = image_height * tf.ones(tf.shape(boxes[:, 2]))
    else:
        horizontal_width = scale * boxes[:, 2]
        vertical_width = scale * boxes[:, 3]

    h_x, h_y, h_w, h_h = centers[:, 0]-horizontal_width, centers[:, 1]-height/2, horizontal_width*2, \
                         tf.ones(tf.shape(centers)[0], tf.float32)*height
    h_x, h_w = _box_crop(h_x, h_w, image_width)
    h_y, h_h = _box_crop(h_y, h_h, image_height)

    horizontal_boxes = tf.stack([h_x, h_y, h_w, h_h], axis=1)

    v_x, v_y, v_w, v_h = centers[:, 0]-width/2, centers[:, 1]-vertical_width, tf.ones(tf.shape(centers)[0])*width, \
                         vertical_width*2
    v_x, v_w = _box_crop(v_x, v_w, image_width)
    v_y, v_h = _box_crop(v_y, v_h, image_height)

    vertical_boxes = tf.stack([v_x, v_y, v_w, v_h], axis=1)

    horizontal_boxes = tf.expand_dims(horizontal_boxes, axis=1)
    vertical_boxes = tf.expand_dims(vertical_boxes, axis=1)

    return tf.concat([horizontal_boxes, vertical_boxes], axis=1)



import tensorflow as tf
import tensorflow.keras.backend as kb
from typing import List, Tuple

from histomics_detect.metrics.iou import iou, greedy_iou_mapping


def normal_loss(loss_object: tf.keras.losses.Loss, boxes: tf.Tensor, rpn_boxes_positive: tf.Tensor,
                scores: tf.Tensor, positive_weight: float, standard: List[tf.keras.metrics.Metric] = [],
                weighted_loss: bool = False, neg_pos_loss: bool = False, use_pos_neg_loss: bool = False,
                min_iou: float = 0.18) \
        -> Tuple[tf.Tensor, tf.Tensor]:
    """
    calculates the normal loss of a lnms output

    labels are calculated based on the largest iou, the prediction that is closest to the respective
    ground truth gets assigned a 1 label and the rest a 0

    then a loss is applied to the objectiveness score output 'nms_output' and the labels

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    loss_object:
        loss function for loss calculation between 'labels' and 'nms_output'
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_boxes_positive: tensor (float32)
        predicted boxes
        shape: N x 4
    scores: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    positive_weight: float
        weight applied to the positive labels ( == 1)
    standard: metric
        list of tensorflow metrics
        1, 2 should be positive and negative loss respectively if 'neg_pos_loss' set to true
    weighted_loss: bool
        if true, loss of positive labels is weighted by the difference in numbers of positive and negative
        labels
    neg_pos_loss: bool
        if true, the loss of the positive and the negative labels is calculated and logged in the metrics
    use_pos_neg_loss: bool
        returns the weighted sum of the pos and neg loss instead of the normal loss
        !!! only works if neg_pos_loss is also true
    min_iou: float
        minimum iou such that box is considered positive prediction

    Returns
    -------
    loss: float
        loss value
    indexes: tensor (float32)
        indexes of the values that correspond to positive anchors
    """
    labels, indexes = calculate_labels(boxes, rpn_boxes_positive, tf.shape(scores), min_iou)

    # calculate negative and positive labels loss for comparing experiment
    if neg_pos_loss:
        (pos_loss, neg_loss), (positive_labels, negative_labels) = _pos_neg_loss_calculation(scores, labels,
                                                                                             loss_object, standard)
        # use negative or positive for training model
        if use_pos_neg_loss:
            return pos_loss * positive_weight + neg_loss, indexes

    # weigh loss
    if weighted_loss:
        num_pos = tf.cast(tf.size(positive_labels), tf.float32)
        num_neg = tf.cast(tf.size(negative_labels), tf.float32)
        weighted_labels = tf.cast(labels, tf.float32) * num_neg / num_pos * positive_weight
        weight = weighted_labels + (1 - labels)

        loss = loss_object(weighted_labels, scores * weight)
    else:
        loss = loss_object(labels, scores)

    return tf.reduce_sum(loss), indexes


def paper_loss(boxes: tf.Tensor, rpn_boxes_positive: tf.Tensor, nms_output: tf.Tensor,
               loss_object: tf.keras.losses.Loss, positive_weight: float, standard: List[tf.keras.metrics.Metric],
               weighted_loss: bool = False, neg_pos_loss: bool = False, min_iou: float = 0.18) \
        -> Tuple[tf.Tensor, tf.Tensor]:
    """
    loss calculation of the paper "Learning Non-Max Suppression"

    the loss is calculated with:
    - the labels vector l with 1s for positive labels and -1 for negative labels
    - the score output of the network n with values btw -1 and 1

    - calculation: positive_label_weight * log(1 + exp(-l * n))

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    loss_object:
        loss function for loss calculation between 'labels' and 'nms_output'
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_boxes_positive: tensor (float32)
        predicted boxes
        shape: N x 4
    nms_output: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    positive_weight: float
        weight applied to the positive labels ( == 1)
    standard: metric
        list of tensorflow metrics
        1, 2 should be positive and negative loss respectively if 'neg_pos_loss' set to true
    weighted_loss: bool
        if true, loss of positive labels is weighted by the difference in numbers of positive and negative
        labels
    neg_pos_loss: bool
        if true, the loss of the positive and the negative labels is calculated and logged in the metrixes
    min_iou: float
        minimum iou such that box is considered positive prediction

    Returns
    -------
    loss: float
        loss value
    indexes: tensor (float32)
        indexes of the values that correspond to positive anchors

    """
    labels, indexes = calculate_labels(boxes, rpn_boxes_positive, tf.shape(nms_output), min_iou)

    # calculate pos and neg loss
    if weighted_loss or neg_pos_loss:
        _, (positive_labels, negative_labels) = _pos_neg_loss_calculation(nms_output, labels, loss_object, standard)

    if weighted_loss:
        num_pos = tf.cast(tf.size(positive_labels), tf.float32)
        num_neg = tf.cast(tf.size(negative_labels), tf.float32)

        weight = labels * num_neg / (num_pos + 1e-8) * positive_weight + (1 - labels)
    else:
        weight = tf.ones(tf.shape(nms_output))

    # reformat labels and output from 0, 1 space to -1, 1 space
    labels = (2 * labels - 1)
    nms_output = (2 * nms_output) - 1

    # calculate loss
    loss = weight * kb.log(1 + kb.exp(-labels * nms_output))
    loss = tf.reduce_sum(loss)
    return loss, indexes


def calculate_labels(boxes, rpn_boxes_positive, output_shape, min_iou: float = 0.18):
    """
    calculate the labels for the predictions
    each ground truth has one positive predictions (label = 1) and the other predictions are
    negative (label = 0)

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_boxes_positive: tensor (float32)
        predicted boxes
        shape: N x 4
    output_shape: tensor (int32)
        shape of the label output
    min_iou: float
        minimum iou such that box is considered positive prediction

    Returns
    -------
    labels: tensor (int32)
        tensor with one entry per prediction
        1 -> if prediction is corresponding to a ground truth
        0 -> if prediction is not corresponding to a ground truth
    indexes: tensor (int32)
        indexes of the predictions that are positive
    """
    ious = iou(rpn_boxes_positive, boxes)

    tp, fp, fn, tp_list, fp_list, fn_list = greedy_iou_mapping(ious, min_iou)

    indexes = tf.reshape(tp_list[:, 0], (-1, 1))
    labels = tf.scatter_nd(indexes, tf.ones(tf.shape(indexes)), output_shape)

    # ious, _ = iou(boxes, rpn_boxes_positive)

    # function that finds prediction with highest overlap with ground truth
    # def assignment_func(i) -> tf.int32:
    #     index = tf.cast(i, tf.int32)
    #     assignment = tf.cast(tf.argmax(ious[index]), tf.int32)
    #     return assignment
    #
    # indexes = tf.map_fn(lambda x: assignment_func(x), tf.range(0, tf.shape(ious)[0]))
    # indexes = tf.expand_dims(indexes, axis=1)
    # labels = tf.scatter_nd(indexes, tf.ones(tf.shape(indexes)), output_shape)

    return labels, indexes


def _pos_neg_loss_calculation(nms_output: tf.Tensor, labels: tf.Tensor, loss_object: tf.keras.losses.Loss,
                              standard: List[tf.keras.metrics.Metric]) \
        -> Tuple[Tuple[tf.Tensor, tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]:
    """

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    nms_output: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    labels: tensor (int32)
        ground truth labels of corresponding s
        shape: N x 1
    loss_object:
        loss function for loss calculation between 'labels' and 'nms_output'
    standard: metric
        list of tensorflow metrics
        1, 2 should be positive and negative loss respectively if 'neg_pos_loss' set to true

    Returns
    -------
    pos_loss: tensor (float32)
        scalar value
    neg_loss: tensor (float32)
        scalar value
    positive_labels: tensor (int32)
        ones for the number of positive ground truth samples
    negative_labels:
        zeros for the number of positive ground truth samples

    """

    positive_predictions, negative_predictions = tf.dynamic_partition(nms_output, tf.cast(labels == 0, tf.int32), 2)
    positive_labels = tf.ones(tf.shape(positive_predictions))
    negative_labels = tf.zeros(tf.shape(negative_predictions))

    # calculate loss
    pos_loss = tf.reduce_sum(loss_object(positive_predictions, positive_labels))
    neg_loss = tf.reduce_sum(loss_object(negative_predictions, negative_labels))

    zero_func = lambda: 0.0
    pos_loss = tf.cond(tf.size(positive_labels) > 0, lambda: pos_loss, zero_func)
    neg_loss = tf.cond(tf.size(negative_labels) > 0, lambda: neg_loss, zero_func)

    # update metrics
    standard[1].update_state(pos_loss + 1e-8)
    standard[2].update_state(neg_loss + 1e-8)

    return (pos_loss, neg_loss), (positive_labels, negative_labels)


def cluster_labels_indexes(scores, cluster_assignment) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    calculate the labels for the predictions based on clusters and scores

    the 'cluster_assignment' relates predictions to a cluster for each ground truth
    for each cluster the prediction with the highest score is assigned a positive label (label = 1)
    the rest is assigned a negative label (label = 0)

    N: number of predictions

    Parameters
    ----------
    scores: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    cluster_assignment: tensor (int32)
        cluster labels for each prediction
        shape: N x 1

    Returns
    -------

    """
    cluster_assignment = tf.expand_dims(cluster_assignment, axis=1)

    # find prediction index with highest objectiveness in cluster
    def max_cluster_index_func(i) -> tf.int32:
        index = tf.cast(i, tf.int32)
        max_index = tf.argmax(
            tf.multiply(tf.cast(scores, tf.float32),
                        tf.cast(tf.equal(cluster_assignment, tf.cast(index, tf.int32)),
                                tf.float32)))
        return tf.cast(max_index, tf.int32)

    indexes = tf.map_fn(lambda x: max_cluster_index_func(x), tf.range(0, tf.reduce_max(cluster_assignment) + 1))

    labels = tf.scatter_nd(indexes, tf.ones(tf.shape(indexes)), tf.shape(scores))

    return labels, indexes


def clustering_loss(nms_output: tf.Tensor, cluster_assignment: tf.Tensor, loss_object: tf.keras.losses.Loss,
                    positive_weight: float, standard: List[tf.keras.metrics.Metric], boxes: tf.Tensor,
                    rpn_positive: tf.Tensor, weighted_loss: bool = False, neg_pos_loss: bool = False,
                    add_regression_param: int = 0) \
        -> Tuple[tf.Tensor, tf.Tensor]:
    """
    clustering loss calculation

    the loss is calculated by:
    - for each cluster the prediction with the highest objectiveness score is stored
    - the index of the stored predictions is set to one in a labels vector
    - the values of the other indexes are 0
    - the loss is calculated by calculating the difference btw. the labels and the nms_output

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    G: number of ground truth boxes

    Parameters
    ----------
    loss_object:
        loss function for loss calculation between 'labels' and 'nms_output'
    nms_output: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    cluster_assignment: tensor (int32)
        cluster labels for each prediction
        shape: N x 1
    positive_weight: float
        weight applied to the positive labels ( == 1)
    standard: metric
        list of tensorflow metrics
        1, 2 should be positive and negative loss respectively if 'neg_pos_loss' set to true
    boxes: tensor (int32)
        ground truth boxes
    rpn_positive: tensor (float32)
        predicted boxes
    weighted_loss: bool
        if true, loss of positive labels is weighted by the difference in numbers of positive and negative
        labels
    neg_pos_loss: bool
        if true, the loss of the positive and the negative labels is calculated and logged in the metrics
    add_regression_param: int
        0 -> lnms only predicts a single obj. score
        1 -> lnms also regresses the center of the boxes
        2 -> lnms regresses the full boxes
    # TODO add weighting for regression vs score loss
    Returns
    -------
    loss: float
        loss value
    indexes: tensor (float32)
        indexes of the values that correspond to positive anchors
    """
    scores = tf.expand_dims(nms_output[:, 0], axis=1)

    labels, indeces = cluster_labels_indexes(scores, cluster_assignment)

    # calculate pos and neg loss
    if neg_pos_loss:
        _pos_neg_loss_calculation(scores, labels, loss_object, standard)

    if weighted_loss:
        weight = labels * positive_weight + (1 - labels)
    else:
        weight = tf.ones(tf.shape(scores))

    if add_regression_param > 0:
        reg = nms_output[:, 1:add_regression_param * 2 + 1]

        def pos_prediction_dist_func(i) -> tf.float32:
            index = tf.cast(i, tf.int32)
            cluster_scores = tf.multiply(tf.cast(scores, tf.float32),
                                         tf.cast(tf.equal(cluster_assignment, tf.cast(index, tf.int32)),
                                                 tf.float32))
            max_index = tf.cast(tf.argmax(cluster_scores), tf.int32)[0]

            dist = tf.math.sigmoid(
                (boxes[index, :add_regression_param * 2] - rpn_positive[max_index, :add_regression_param * 2]) / 100)

            return tf.cast(dist, tf.float32)

        distances = tf.map_fn(lambda x: pos_prediction_dist_func(x),
                              tf.cast(tf.range(0, tf.reduce_max(cluster_assignment) + 1), tf.float32))

        distance_vector = tf.scatter_nd(indeces, distances, tf.shape(reg))

        loss_score = loss_object(weight * labels, weight * scores)
        loss_reg = loss_object(distance_vector, labels * reg)

        return tf.reduce_sum(loss_score + loss_reg), labels
    else:
        loss = loss_object(weight * labels, weight * scores)
        return tf.reduce_sum(loss), labels


def normal_clustering_loss(nms_output: tf.Tensor, boxes: tf.Tensor, rpn_boxes_positive: tf.Tensor,
                           cluster_assignment: tf.Tensor, loss_object: tf.keras.losses.Loss,
                           positive_weight: float, standard: List[tf.keras.metrics.Metric], weighted_loss: bool = False,
                           neg_pos_loss: bool = False, use_pos_neg_loss: bool = False, norm_loss_weight: float = 1,
                           add_regression_param: int = 0, min_iou: float = 0.18) -> Tuple[float, tf.Tensor]:
    """
    a combination between the normal and clustering loss

    loss = 'norm_loss_weight' * normal_loss + clustering_loss

    Parameters
    ----------
    nms_output: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_boxes_positive: tensor (float32)
        predicted boxes
        shape: N x 4
    cluster_assignment: tensor (int32)
        cluster labels for each prediction
        shape: N x 1
    loss_object:
        loss function for loss calculation between 'labels' and 'nms_output'
    positive_weight: float
        weight applied to the positive labels ( == 1)
    standard: metric
        list of tensorflow metrics
        1, 2 should be positive and negative loss respectively if 'neg_pos_loss' set to true
    weighted_loss: bool
        if true, loss of positive labels is weighted by the difference in numbers of positive and negative
        labels
    neg_pos_loss: bool
        if true, the loss of the positive and the negative labels is calculated and logged in the metrics
    use_pos_neg_loss: bool
        returns the weighted sum of the pos and neg loss instead of the normal loss
        !!! only works if neg_pos_loss is also true
    norm_loss_weight: float
        weight of the normal loss
    add_regression_param: int
        0 -> lnms only predicts a single obj. score
        1 -> lnms also regresses the center of the boxes
        2 -> lnms regresses the full boxes
    min_iou: float
        minimum iou such that box is considered positive prediction
    # TODO add weighting for regression vs score loss

    Returns
    -------
    loss: float
        the combined loss
    indexes: tensor (float32)
        indexes of the values that correspond to positive anchors
    """
    scores = tf.expand_dims(nms_output[:, 0], axis=1)
    norm_loss, indexes = normal_loss(loss_object, boxes, rpn_boxes_positive, scores, positive_weight, standard,
                                     weighted_loss, neg_pos_loss, use_pos_neg_loss, min_iou)
    clust_loss, _ = clustering_loss(nms_output, cluster_assignment, loss_object, positive_weight, standard, boxes,
                                    rpn_boxes_positive, weighted_loss, neg_pos_loss, add_regression_param)

    loss = norm_loss_weight * norm_loss + clust_loss

    return loss, indexes


def xor_loss(nms_output: tf.Tensor, cluster_assignment: tf.Tensor):
    """
    xor loss
    the loss is minimal if only one score of each cluster is one and the others are zero

    calculation for each cluster:
    - calculate cluster sum
    - subtract one and square result

    calculate for each prediction
    - subtract 1/2 from the score
    - square the result
    - subtract from previous result

    sum over all prediction losses

    Parameters
    ----------
    nms_output: tensor (float32)
        output scores for each prediction
    cluster_assignment: tensor (int32)
        assignment of each prediction to the corresponding cluster

    Returns
    -------
    loss: float
        calculated loss
    """

    # TODO find error cause
    # TODO add optional neg pos loss calculation

    def cluster_sum(i) -> tf.float32:
        pred_indexes = tf.where(tf.equal(tf.cast(cluster_assignment, tf.float32), tf.cast(i, tf.float32)))
        predictions = tf.gather_nd(nms_output, pred_indexes)

        sum_req = (tf.reduce_sum(predictions) - 1) ** 2

        indexes = tf.cast(pred_indexes, tf.int64)
        update_shape = tf.cast(tf.shape(cluster_assignment), tf.int64)

        false_fn = lambda: tf.scatter_nd(indexes, tf.ones(tf.shape(indexes)[0]) * sum_req, update_shape)
        scattered_sum = tf.cond(tf.size(indexes) == 0, lambda: tf.zeros(update_shape), false_fn)
        return tf.squeeze(scattered_sum)

    number_clusters = tf.reduce_max(cluster_assignment) + 1
    number_predictions = tf.shape(cluster_assignment)[0]

    output_signature = tf.TensorSpec.from_tensor(tf.ones(number_predictions, dtype=tf.float32))

    cluster_sums = tf.map_fn(lambda x: cluster_sum(x), tf.range(0, number_clusters), dtype=output_signature)
    cluster_sums = tf.expand_dims(tf.reduce_sum(cluster_sums, axis=0), axis=1)

    loss = tf.reduce_sum((cluster_sums - 1) ** 2 - (nms_output - 0.5) ** 2, axis=0)
    return loss, None



from typing import Tuple
from histomics_detect.boxes import filter_edge_boxes
import tensorflow as tf


@tf.function
def create_anchors(anchor_px, field, width, height, filter_boxes=True):
    """Generates anchors given anchor sizes, receptive field size, and input image size.
    
    This function generates the complete set of anchors for an image with dimensions 
    height x width and backbone with given receptive field size, and removes anchors
    that cross the image boundary. Anchors are squares whose centers coincide with the 
    centers of the receptive fields of the backbone network. Each receptive field 
    corresponds to a feature map generated by the backbone, and each field has a set of
    anchors of various sizes. Each size corresponds to an output from the objectness and 
    regression branches of the region proposal network.
        
    Parameters
    ----------
    anchor_px: tensor (int32)
        K-length 1-d tensor containing the anchor width hyperparameter values in pixel 
        units.
    field: float32
        Edge length of the receptive field in pixels. This defines the area of the 
        image that corresponds to 1 feature map and the anchor gridding.
    width: int32
        Input image width in pixels.
    height: int32
        Input image height in pixels.
    filter_boxes: bool
        if true boxes are filtered, if they overlap with the border
        
    Returns
    -------
    anchors: tensor (float32)
        M x 4 tensor anchor positions organized in a dense grid over the image
        space. Each row contains the x,y upper left corner of the anchor in pixel 
        units relative in the image coordinate frame, and the anchor width and 
        height.
    """

    # generate unfiltered anchor locations using smallest anchors size
    px = tf.expand_dims(anchor_px, 1)

    x_pair, y_pair = _generate_x_y_pairs(px, field, width, height)

    # transform from 2D meshgrid format to array format (1 row per anchor)
    single_size_anchors = tf.concat((tf.reshape(x_pair, (tf.size(x_pair), 1)),
                                    tf.reshape(y_pair, (tf.size(x_pair), 1)),
                                    tf.ones((tf.size(x_pair), 1)),
                                    tf.ones((tf.size(x_pair), 1))),
                                    axis=1)

    # replicate array for each anchor size
    anchors = tf.tile(tf.expand_dims(single_size_anchors, axis=0), [tf.size(px), 1, 1])

    # add the width and height to the anchors
    px_expanded = tf.cast(tf.expand_dims(px, axis=1), tf.float32)
    ones = tf.ones((tf.size(px), 1, 1), tf.float32)
    multiplier = tf.concat((ones, ones, px_expanded, px_expanded), axis=2)
    anchors = tf.reshape(anchors * multiplier, (-1, 4))

    # remove anchors that cross the boundary
    if filter_boxes:
        anchors, _ = filter_edge_boxes(anchors, width, height, 0, tf.constant(False, tf.bool))


    return anchors


@tf.function
def _generate_x_y_pairs(px, field, width, height):
    """Generates 2D tensors containing the (x,y) corner positions of anchors for a
    square px x px anchor, height x width image, and given receptive field size.
    
    Each receptive field has a corresponding set of anchors with varying sizes.
    Since anchors must be contained entirely within the image border, some receptive
    fields at the border may lack a complete set of anchors. This function generates
    the complete set of anchor positions for a given anchor size in a 2D format similar
    to 'meshgrid'. The output includes anchors that cross the image border (these 
    anchors are later filtered using box transformations).
        
    Parameters
    ----------
    size: int32
        Anchor size in pixels.
    field: float32
        Edge length of the receptive field in pixels. This defines the area of the 
        image that corresponds to 1 feature map from the backbone network grid spacing
        of anchors.
    width: int32
        Image width in pixels.
    height: int32
        Image height in pixels.
        
    Returns
    -------
    x_pair: tensor (float32)
        A 2D grid of anchor corner horizontal positions. Each position in the grid
        represents a receptive field. Includes anchors that cross image boundaries.
    y_pair: tensor (float32)
        A 2D grid of anchor corner vertical positions. Each position in the grid
        represents a receptive field. Includes anchors that cross image boundaries.
    """    
    
    # the smallest anchor size will determine spacing from the border - larger 
    # anchor sizes will be filtered later
    size = tf.reduce_min(px)

    # get anchor corners
    x_corners = _anchor_corners(size, field, width)
    y_corners = _anchor_corners(size, field, height)

    # replicate for coordinate pairing
    x_pair = tf.tile(tf.expand_dims(x_corners, axis=0), [tf.size(y_corners), 1])
    y_pair = tf.tile(tf.expand_dims(y_corners, axis=1), [1, tf.size(x_corners)])

    return x_pair, y_pair


def first_last_anchor_indexes(size, field, length) -> Tuple[int, int]:
    """
    calculates the first and last indexes of the anchors where a box of the size 'size' does not intersect
    with the boundary of the image

    Parameters
    ----------
    size: int32
        Anchor size.
    field: float32
        Edge length of the receptive field in pixels. This defines the area of the
        image that corresponds to 1 feature map from the backbone network and the
        anchor gridding.
    length: int32
        Image edge length in pixels.

    Returns
    -------
    first: int32
        index of the first anchor that does not intersect with the boundary
    last: int32
        index of the last anchor
    """
    first = tf.math.ceil(tf.cast(size, tf.float32) /
                         (2 * tf.cast(field, tf.float32)) - 1 / 2)
    last = tf.math.floor((tf.cast(length, tf.float32) -
                          tf.cast(size, tf.float32) / 2) /
                         tf.cast(field, tf.float32) - 1 / 2)

    return first, last


@tf.function
def _anchor_corners(size, field, length):
    """Generates a sequence of anchor corner positions along one dimension of an image.
    
    Anchors are required to be entirely with the image domain during training. Called
    twice for each image to generate a horizontal sequence and a vertical sequence.
    These sequences are combined using a cartesian product to generate a dense grid
    of anchors over the entire image.
        
    Parameters
    ----------
    size: int32
        Anchor size.
    field: float32
        Edge length of the receptive field in pixels. This defines the area of the 
        image that corresponds to 1 feature map from the backbone network and the 
        anchor gridding.
    length: int32
        Image edge length in pixels.
        
        
    Returns
    -------
    corners: tensor (float32)
        1D tensor of upper/left anchor corner positions in pixels.
    """

    # first and last anchor index
    first, last = first_last_anchor_indexes(size, field, length)

    # anchor corners
    corners = tf.cast(field, tf.float32) * (tf.range(first, last + 1) + 1 / 2) - tf.cast(size, tf.float32) / 2

    return corners



import tensorflow as tf


@tf.function
def sample_anchors(positive, negative, negative_obj, max_anchors=256, np_ratio=2.0,
                   hard_fraction=0.0):
    """Sample to balance the proportion of positive and negative anchors used in 
    training.
    
    Samples up to max_n anchors with a maximum negative : positive ratio from the
    dense anchors in the image space.
        
    Parameters
    ----------
    positive: tensor (float32)
        M x 4 tensor of positive anchor positions satisfying IoU criteria
        with some ground truth object. Each row contains the x,y center location 
        of the anchor in pixel units relative in the image coordinate frame, and
        the anchor width and height.
    negative: tensor (float32)
        N x 4 tensor of negative anchor positions not satisfying IoU criteria with
        any ground truth object. Each row contains the x,y center location of the
        anchor in pixel units relative in the image coordinate frame, and the
        anchor width and height.
    negative_obj: tensor (float32)
        N or N x 1 tensor of objectness scores for negative anchors. This is used
        to enrich the sampled negative anchors with hard negatives by selecting
        misclassified negatives with high scores.
    max_anchors: int32
        Maximum number of total anchors to sample. Default value 256.
    np_ratio: float32
        Will sample at most negative : positive ratio anchors. Default value 2.0.
    hard_fraction: float32
        Make hard_fraction percent of negative anchors consist of hard negatives
        that are missclassified. Range is [0, 1]. Default value 0.
        
    Returns
    -------
    positive: tensor (float32)
        m x 4 tensor of positions of positive sampled anchors, where m <= M.
    negative: tensor (float32)
        n x 4 tensor of positions of negative sampled anchors, where n <= N and
        n + m <= max_anchors and n : m < np_ratio.
    """

    #calculate negative anchor limit if positive anchors abundant
    neg_lim = max_anchors - tf.cast(tf.cast(max_anchors / (1 + np_ratio), tf.float32), tf.int32)
    nneg = tf.minimum(tf.cast(tf.cast(tf.shape(positive)[0], tf.float32) * np_ratio, tf.int32),
                      neg_lim)
    nneg = tf.minimum(tf.shape(negative)[0], nneg)
    nneg = tf.maximum(nneg, 1)
    
    #sample positive anchors
    npos = tf.minimum(tf.shape(positive)[0], max_anchors-nneg)
    indices = _sample_no_replacement(tf.shape(positive)[0]-1, npos)
    positive = tf.gather(positive, indices)
    
    #take the top ceil(hard_frac * nneg) hardest negatives
    nhard = tf.cast(tf.math.ceil(hard_fraction * tf.cast(nneg, tf.float32)), tf.int32)
    
    #sort negative anchors by objectness score (descending)
    order = tf.argsort(negative_obj, axis=0, direction='DESCENDING')
    hard_indices, other_indices = tf.split(order, [nhard, tf.shape(negative)[0]-nhard])
    
    #get hard anchors
    hard = tf.gather(negative, hard_indices, axis=0)
    
    #sample remaining negative anchors
    random = _sample_no_replacement(tf.shape(negative)[0]-nhard-1, nneg-nhard)
    other = tf.gather(negative, tf.gather(other_indices, random), axis=0)
    
    #combine hard negatives and randomly sampled negatives
    negative = tf.concat([hard, other], axis=0)
    
    return positive, negative

 
@tf.function
def _sample_no_replacement(maxval, size):
    """Generates indices to sample from a tensor without replacement.
    
    This function generates a maxval-length tensor of random values
    and uses the sorting indices of this array to generate the random
    sample.
        
    Parameters
    ----------
    maxval: int32
        Value of the largest possible index to sample. 
    size: int32
        The number of elements to sample.
        
    Returns
    -------
    positive: tensor (float32)
        m x 4 tensor of positions of positive sampled anchors, where m <= M.
    negative: tensor (float32)
        n x 4 tensor of positions of negative sampled anchors, where n <= N and
        n + m <= max_n and m/n >= min_ratio.
    """
    
    #sample 'size' int32 values up to 'maxval' without replacement
    z = tf.random.uniform((maxval+1,), 0, 1)
    _, indices = tf.nn.top_k(z, size)
    
    return indices



import matplotlib.pyplot as plt
import tensorflow as tf

plt.rcParams['figure.dpi'] = 300 #set figure dots-per-inch


def _plot_boxes(boxes, color):
    """Adds bounding boxes with chosen color to current plot.
        
    Parameters
    ----------
    boxes: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of a regressed box and its width and height in that order.
    color:
        Color argument for matplotlib pyplot.
    """    
    
    x, y, w, h = tf.split(boxes, 4, axis=1)
    for i, (xi, yi, wi, hi) in enumerate(zip(x, y, w, h)):
        plt.plot([xi, xi+wi, xi+wi, xi, xi],
                 [yi, yi, yi+hi, yi+hi, yi],
                 color=color)

        
def plot_inference(rgb, regressions, color='g'):
    """Generates plot of inference results.
        
    Parameters
    ----------
    rgb: array
        Image for display with imshow.
    regressions: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of each regressed box and its width and height in that order.
    color:
        Color argument for matplotlib pyplot.    
        
    See also
    --------
    greedy_iou, _plot_boxes
    """    
    
    plt.imshow(rgb)
    _plot_boxes(regressions, color)
        
    
def plot_evaluation(rgb, boxes, regressions, tp_list, fp_list, fn_list):
    """Generates color coded plot of evaluation results where true positives,
    false positives, and false negatives are color coded.
        
    Parameters
    ----------
    rgb: array
        Image for display with imshow.
    boxes: tensor
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of each ground truth box and its width and height in that order.
    regressions: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of each regressed box and its width and height in that order.
    tp_list: tensor
        A 2-d tensor containing indices of true positive predictions in regressions
        in the first column and the corresponding indices of matched ground truth 
        in boxes.
    fp_list: tensor
        A 1-d tensor of indices of false positive predictions in regressions.
    fn_list: tensor
        A 1-d tensor of indices of false negatives in regressions as produced by
        the function greedy_iou.
        
    See also
    --------
    greedy_iou, _plot_boxes
    """    
    
    plt.imshow(rgb)
    _plot_boxes(tf.gather(regressions, tp_list[:,0], axis=0), 'g')
    _plot_boxes(tf.gather(regressions, fp_list, axis=0), 'r')
    _plot_boxes(tf.gather(boxes, fn_list, axis=0), 'b')



from abc import ABC
from typing import Tuple, List
import tensorflow as tf

from histomics_detect.boxes.neighborhood import all_neighborhoods_additional_info


class BlockModel(tf.keras.Model, ABC):
    def __init__(self, blocks: List[Tuple[tf.keras.Model, tf.keras.Model]], final_layers: tf.keras.Model,
                 threshold: float = 0.5, train_tile: float = 224, use_image_features: bool = True,
                 use_distance: bool = False, original_lnms: bool = True):
        """
        Learning-NMS block model

        Parameters
        ----------
        blocks: List[Tuple[tf.keras.Model, tf.keras.Model]]
            list of blocks each consists of two keras model
            - block: layers, which process the neighborhoods representation
            - output: layers, which process a neighborhood back into a single representation
        final_layers: tf.keras.Model
            layers that compress the final prediction representations into the final output score
        threshold: float
            threshold for two predictions to belong in the same neighborhood
        train_tile: float
            size of the image tile used for training
        use_image_features: bool
            true use image features when creating prediction representation
        use_distance: bool
            assemble neighborhood with distance instead of iou
        original_lnms: bool
            reuse the learned representation for next block
        """
        super(BlockModel, self).__init__(name='block_model')
        self.blocks = blocks
        self.final_layers = final_layers

        self.threshold = threshold
        self.train_tile = train_tile
        self.use_image_features = use_image_features
        self.use_distance = use_distance
        self.original_lnms = original_lnms

    def call(self, x: Tuple[tf.Tensor, tf.Tensor], training: bool = False, mask=None):
        """
        executes a forward pass of the block model for the given data 'x'

        S: size of neighborhood
        N: number of predictions
        D: size of a single prediction

        Parameters
        ----------
        x: (tensor, tensor) (float32)
            (interpolated, rpn_boxes)
            interpolated: interpolated representation of each rpn_box + the corresponding objectiveness score
                        shape (N, ...+1)
            rpn_boxes: boxes
                       shape (N, 4)
        training: bool
            true while training
            false during inference
        mask:
            not used

        Returns
        -------
        nms_output: tensor (float32)
            an updated score (objectiveness) for each rpn_box

        """

        interpolated, rpn_boxes_positive = x

        # initialize loop paramters
        num_predictions = tf.shape(rpn_boxes_positive)[0]
        prediction_ids = tf.range(0, num_predictions)

        neighborhood_sizes, neighborhoods_add_info, neighborhoods_indexes, self_indexes = \
            all_neighborhoods_additional_info(rpn_boxes_positive, prediction_ids, self.train_tile, self.threshold,
                                              self.use_distance)

        if not self.use_image_features and not self.original_lnms:
            empty_features = tf.zeros((tf.shape(interpolated)[0], tf.shape(interpolated)[1] - 1))
            interpolated = tf.concat([tf.expand_dims(interpolated[:, 0], axis=1), empty_features], axis=1)

        for block, output in self.blocks:
            # run network on block

            # assemble neighborhood
            if self.use_image_features or self.original_lnms:
                # assemble the neighboring prediction representation for each neighborhood
                neighborhood = tf.reshape(tf.gather(interpolated, neighborhoods_indexes),
                                          [-1, tf.shape(interpolated)[1]])
                # assemble the tiled prediction representation of the main prediction of each neighborhood
                main_predictions = tf.reshape(tf.gather(interpolated, self_indexes), [-1, tf.shape(interpolated)[1]])
                neighborhoods = tf.concat([neighborhood, main_predictions, neighborhoods_add_info], axis=1)
            else:
                neighborhood = tf.reshape(tf.gather(interpolated[:, 0], neighborhoods_indexes),
                                          [-1, 1])
                main_predictions = tf.reshape(tf.gather(interpolated[:, 0], self_indexes),
                                              [-1, 1])
                empty_features = tf.zeros((tf.shape(neighborhoods_add_info)[0], tf.shape(interpolated)[1] - 1))
                neighborhoods = tf.concat(
                    [neighborhood, empty_features, main_predictions, empty_features, neighborhoods_add_info], axis=1)
            num_predictions = tf.size(neighborhood_sizes)

            # run block
            processed_neighborhoods = block(neighborhoods, training=training)

            # pool neighborhoods
            pooled_predictions = tf.TensorArray(dtype=tf.float32, size=num_predictions)

            counter = tf.constant(0, dtype=tf.int32)
            start_index = tf.constant(0)

            # pool each neighborhood to one prediction representation
            for size in neighborhood_sizes:
                neighborhood_slice = processed_neighborhoods[start_index:start_index + size]
                pooled_predictions = pooled_predictions.write(counter, tf.math.reduce_max(neighborhood_slice, axis=0))

                counter += 1
                start_index += size

            pooled_predictions = pooled_predictions.stack()

            # run final block layers
            processed_predictions = output(pooled_predictions, training=training)
            block_output = tf.cast(processed_predictions, tf.float32)

            # skip connection
            interpolated = (interpolated + block_output)
        return self.final_layers(interpolated, training=training)
